from typing import Any, Dict, List, Optional, Set, Tuple
import collections
import gc
import itertools
import multiprocessing as mp
import os
import queue
import socket
import time
import uuid

import deepspeed
import numpy as np
import pynvml
import torch
import torch.distributed as dist
import torch.utils.data

from realhf.api.core.config import ModelName
from realhf.base import (
    constants,
    gpu_utils,
    logging,
    namedarray,
    numpy_utils,
    recover,
    seeding,
    timeutil,
    topology,
)
from realhf.base.monitor import (
    cuda_tmark,
    cuda_tmarked,
    CUDATimeMarkType,
    dump_tmark_db,
    gpu_utilization_monitor,
)
from realhf.impl.model.nn.real_llm_api import ReaLModel
from realhf.system import request_reply_stream, worker_base
import realhf.api.core.dfg as dfg
import realhf.api.core.system_api as system_api
import realhf.impl.model.comm.data_transfer as data_transfer_comm
import realhf.impl.model.comm.global_comm as global_comm
import realhf.impl.model.comm.param_realloc as param_realloc_comm

# NOTE: Register all implemented datasets and models.
import realhf.api.core.data_api as data_api  # isort:skip
import realhf.api.core.model_api as model_api  # isort:skip

logger = logging.getLogger("Model Worker", "colored")
blogger = logging.getLogger("benchmark")

TIME_RECORD_RPCS = [
    "generate",
    "inference",
    "train_step",
    "save",
    "evaluate",
    "initialize",
]


def get_pytorch_profiler(with_stack):
    return torch.profiler.profile(
        activities=[
            torch.profiler.ProfilerActivity.CPU,
            torch.profiler.ProfilerActivity.CUDA,
        ],
        record_shapes=True,
        profile_memory=True,
        with_stack=True,
        with_flops=True,
    )


class NoRequestToHandle(Exception):
    pass


class ModelWorker(worker_base.Worker):

    def _configure(self, cfg: system_api.ModelWorker):
        self.config = cfg
        self.model_names = [s.id.model_name for s in cfg.shards]
        self.shard_indices = [
            cfg.model_topos[s.id.model_name].get_rank(
                data=s.id.dp_rank, pipe=s.id.pp_rank, model=s.id.mp_rank
            )
            for s in cfg.shards
        ]

        self.__experiment_name = self.config.worker_info.experiment_name
        self.__trial_name = self.config.worker_info.trial_name

        self.config.model_rpcs, _ = dfg.build_graph(self.config.model_rpcs)
        self.data2required_rpc_names = self.config.model_rpcs[
            0
        ].data2required_rpc_names

        # NOTE: here worker_index is different from peer/ddp rank
        self.__worker_index = cfg.worker_info.worker_index

        torch.backends.cudnn.benchmark = cfg.cudnn_benchmark
        torch.backends.cudnn.deterministic = cfg.cudnn_deterministic

        seeding.set_random_seed(cfg.seed)

        # Reveal DDP identity of this worker to world.
        gpu_utils.reveal_ddp_identity(
            self.__experiment_name, self.__trial_name, self.__worker_index
        )
        self.__dist_env_resolved = False

        self.__clear_cache_frequency = timeutil.FrequencyControl(
            frequency_steps=self.config.cuda_cache_clear_freq
        )

        r = self.config.worker_info

        # log info
        self.__total_time = 0.01

        # recover info
        self.__recover_run = os.environ.get("RECOVER_RUN", "0") == "1"
        self.__recover_info = (
            recover.load_recover_info() if self.__recover_run else None
        )
        self.__recover_states_root = os.path.join(
            constants.RECOVER_ROOT, self.__experiment_name, self.__trial_name
        )
        self.__epoch_since_recover = 0
        self.__recover_first_epoch_done = False

        return r

    @property
    def _mp_rank(self) -> int:
        return constants.model_parallel_rank()

    @property
    def _pp_rank(self) -> int:
        return constants.pipe_parallel_rank()

    @property
    def _dp_rank(self) -> int:
        return constants.data_parallel_rank()

    @property
    def _pp_size(self) -> int:
        return constants.pipe_parallel_world_size()

    @property
    def _mp_size(self) -> int:
        return constants.model_parallel_world_size()

    @property
    def _dp_size(self) -> int:
        return constants.data_parallel_world_size()

    @property
    def _is_dp_head(self) -> bool:
        return self._mp_rank == 0 and self._pp_rank == self._pp_size - 1

    @property
    def _model(self) -> model_api.Model:
        return self.__models[constants.model_name()]

    @property
    def _interface(self) -> model_api.ModelInterface:
        return self.__interfaces[constants.model_name()]

    @property
    def _eval_dataloader(self) -> torch.utils.data.DataLoader:
        return self.__eval_dataloaders[constants.model_name()]

    @property
    def _module(self) -> torch.nn.Module | ReaLModel:
        return self.__unwrapped_models[constants.model_name()]

    @property
    def _backend(self) -> model_api.ModelBackend:
        return self.__backends[constants.model_name()]

    def __lazy_setup(self):
        # Add an additional subscript pattern for source RPCs.
        self.__has_dataset = False
        self.__dataset_dp_size = self.__dataset_dp_rank = 0
        sub_patterns = [s.id for s in self.config.shards]
        src_rpc = [rpc for rpc in self.config.model_rpcs if rpc.is_src][0]
        self.__src_rpc_model_name = src_rpc.model_name
        for s in self.config.shards:
            _pp_size = s.id.topo.get_dim("pipe")
            if not (s.id.mp_rank == 0 and s.id.pp_rank == _pp_size - 1):
                continue
            if src_rpc.model_name == s.id.model_name:
                self.__has_dataset = True
                self.__dataset_dp_size = s.id.topo.get_dim("data")
                self.__dataset_dp_rank = s.id.dp_rank
                sub_patterns.append(f"__data{self.__dataset_dp_rank}__")
                break

        # Build stream connecting with master workers.
        self.__stream = request_reply_stream.make_worker_stream(
            self.config.worker_info,
            idx=self.__worker_index,
        )

        self.__pg_info = global_comm.setup_global_comm(
            expr_name=self.__experiment_name,
            trial_name=self.__trial_name,
            worker_index=self.__worker_index,
            model_topos=self.config.model_topos,
            msid2mwid=self.config.msid2mwid,
        )

        self.__data_transfer_info = data_transfer_comm.setup_data_transfer(
            model_topos=self.config.model_topos,
            msid2mwid=self.config.msid2mwid,
            data_transfer_pairs=self.config.data_transfer_pairs,
        )

        self.__param_realloc_info = param_realloc_comm.setup_param_realloc(
            model_topos=self.config.model_topos,
            msid2mwid=self.config.msid2mwid,
            param_realloc_pairs=self.config.sync_param_pairs,
        )

        # logger.info(f"SetUp Information - Model worker index {self.__worker_index} located at "
        #             f"{socket.gethostname()} GPU {self.__pg_info.local_gpu_id}.")

        deepspeed.init_distributed()
        # self.logger.info("deepspeed init distributed on model worker")
        self.__device = torch.device("cuda:0")

        for model_name_, topo_ in self.config.model_topos.items():
            rpcs = [
                rpc
                for rpc in self.config.model_rpcs
                if rpc.model_name == model_name_
            ]
            assert len(rpcs) >= 1
            is_trainable_model = any(
                [
                    rpc.interface_type == dfg.ModelInterfaceType.TRAIN_STEP
                    for rpc in rpcs
                ]
            )
            param_realloc_comm.set_trainable(model_name_, is_trainable_model)
            constants.set_rank_mapping(
                model_name_, topo_, self.config.msid2mwid
            )
            grid = topology.ParallelGrid(
                topology=topo_,
                rank_mapping=constants.rank_mapping_of_model(model_name_),
                process_group=self.__pg_info.model_groups[model_name_],
            )
            constants.set_grid(model_name_, grid)

        # Set up training dataset for source RPCs.
        if self.__has_dataset:
            datasets = [
                data_api.make_dataset(
                    d,
                    self.config.seed,
                    self.__dataset_dp_rank,
                    self.__dataset_dp_size,
                    self.config.tokenizer_name_or_path,
                    self.config.worker_info.experiment_name,
                    self.config.worker_info.trial_name,
                    cache_root=(
                        None
                        if not self.config.use_dataset_cache
                        else self.config.dataset_cahce_root
                    ),
                )
                for d in self.config.datasets
            ]
            if len(self.config.datasets) == 1:
                self.__dataset = datasets[0]
            else:
                self.__dataset = torch.utils.data.ConcatDataset(datasets)
            self.__dataloader = data_api.make_dataloader(
                self.config.dataloader, self.__dataset
            )
            self.__dataset_n_seqs = 0
            for tmp_sample in self.__dataloader:
                self.__dataset_n_seqs += len(
                    data_api.split_sequences(tmp_sample)
                )

            self.__data_generator = enumerate(self.__dataloader)
            self.__dataset_batch_counter = None

            self.__dataset_epoch = (
                0
                if not self.__recover_run
                else self.__recover_info.recover_start.epoch
            )
            self.__cur_sample = None
            self.__fetched_sample_cache = []

        self.__models: Dict[ModelName, model_api.Model] = dict()
        self.__model_is_handle: Dict[ModelName, bool] = dict()
        self.__interfaces: Dict[ModelName, model_api.ModelInterface] = dict()
        self.__eval_dataloaders: Dict[
            ModelName, torch.utils.data.DataLoader
        ] = dict()

        self.__backends: Dict[ModelName, model_api.ModelBackend] = dict()
        self.__unwrapped_models: Dict[
            ModelName, torch.nn.Module | ReaLModel
        ] = dict()

        self.__backend_initialized: Dict[ModelName, bool] = dict()

        for s in self.config.shards:
            with constants.model_scope(s.id.model_name):
                self.__backend_initialized[s.id.model_name] = False
                tik = time.perf_counter()
                if self.__recover_run:
                    model_path = os.path.join(
                        self.__recover_states_root, "ckpt", s.id.model_name.role
                    )
                    s.model.args["model_path"] = model_path
                    s.model.args["init_critic_from_actor"] = False

                self.__models[s.id.model_name] = model = model_api.make_model(
                    s.model, name=s.id.model_name, device=self.__device
                )
                self.__unwrapped_models[s.id.model_name] = model.module
                if s.should_instantiate:
                    if isinstance(model.module, ReaLModel):
                        model.module.instantiate()
                    self.__model_is_handle[s.id.model_name] = False
                else:
                    self.__model_is_handle[s.id.model_name] = True
                self.__backends[s.id.model_name] = model_api.make_backend(
                    s.backend
                )
                interface_impl = [
                    rpc.interface_impl
                    for rpc in self.config.model_rpcs
                    if rpc.model_name == s.id.model_name
                ]
                assert all(x == interface_impl[0] for x in interface_impl)
                self.__interfaces[s.id.model_name] = model_api.make_interface(
                    interface_impl[0]
                )

                if (
                    s.eval_datasets is not None
                    and s.eval_dataloader is not None
                ):
                    eval_datasets = [
                        data_api.make_dataset(
                            d,
                            self.config.seed,
                            s.id.dp_rank,
                            s.id.topo.get_dim("data"),
                            self.__models[s.id.model_name].tokenizer,
                            self.config.worker_info.experiment_name,
                            self.config.worker_info.trial_name,
                            cache_root=(
                                None
                                if not self.config.use_dataset_cache
                                else self.config.dataset_cahce_root
                            ),
                        )
                        for d in s.eval_datasets
                    ]
                    if len(eval_datasets) > 1:
                        eval_dataset = torch.utils.data.ConcatDataset(
                            eval_datasets
                        )
                    else:
                        eval_dataset = eval_datasets[0]
                    eval_dataloader = data_api.make_dataloader(
                        s.eval_dataloader, eval_dataset
                    )
                else:
                    eval_dataloader = None
                self.__eval_dataloaders[s.id.model_name] = eval_dataloader

        self.__request_cache = []
        self.__ack_cache = {}

        self.__request_queue = queue.Queue(maxsize=8)
        self.__reply_queue = queue.Queue(maxsize=8)
        self.__request_sample_size = dict()

        # Storing model function call outputs. Model worker will serve as
        # the data producer when other model workers require this data.
        self.__data_owner_storage: Dict[int, Dict[str, torch.Tensor]] = (
            collections.defaultdict(dict)
        )
        self.__data_receive_cache: Dict[int, Dict[str, torch.Tensor]] = (
            collections.defaultdict(dict)
        )

        self.__data_sent_worker_indices: Dict[int, Dict[str, Set]] = (
            collections.defaultdict(lambda: collections.defaultdict(set))
        )
        self.__data_received_worker_indices: Dict[int, Dict[str, Set]] = (
            collections.defaultdict(lambda: collections.defaultdict(set))
        )

        self.__compute_input_queues = {
            model_name: dict(
                train_step=queue.Queue(4),
                inference=queue.Queue(4),
                generate=queue.Queue(4),
                evaluate=queue.Queue(4),
            )
            for model_name in self.__models.keys()
        }

        # A monitoring process.
        self.__gpu_util_mp = mp.Process(
            target=gpu_utilization_monitor, args=(self.__worker_index, 20, 7200)
        )
        # self.__gpu_util_mp.start()

    def __prefetch_from_dataset(self):
        if self.__cur_sample is None:
            try:
                self.__dataset_batch_counter, self.__cur_sample = next(
                    self.__data_generator
                )
            except StopIteration:
                self.__dataset_epoch += 1
                if self.__recover_run:
                    self.__epoch_since_recover += 1
                    if self.__epoch_since_recover > 1:
                        self.__recover_first_epoch_done = True
                self.__data_generator = enumerate(self.__dataloader)
                self.__dataset_batch_counter, self.__cur_sample = next(
                    self.__data_generator
                )

    def __handle_one_rpc_hook(self, hook: str, hook_data: Any):
        if hook == "data_transfer":
            self.__data_transfer_among_workers(hook_data)
        elif hook == "param_realloc":
            from_model_name: ModelName = hook_data["from_model_name"]
            to_model_name: ModelName = hook_data["to_model_name"]
            from_topo: topology.PipeModelDataParallelTopology = hook_data[
                "from_topo"
            ]
            to_topo: topology.PipeModelDataParallelTopology = hook_data[
                "to_topo"
            ]
            to_model_config = hook_data["to_model_config"]
            if from_model_name in self.__unwrapped_models:
                m = self.__unwrapped_models[from_model_name]
            else:
                m = self.__unwrapped_models[to_model_name]
            if not isinstance(m, ReaLModel):
                raise ValueError(
                    f"Model {from_model_name} (type={type(m)}) is not a ReaLModel, "
                    f"so it can't use parameter realloction."
                )
            try:
                new_layers, new_param, _ = m.build_reparallelized_layers_async(
                    from_model_name=from_model_name,
                    to_model_name=to_model_name,
                    from_topo=from_topo,
                    to_topo=to_topo,
                    to_model_config=to_model_config,
                    pg_info=self.__param_realloc_info,
                )
            except RuntimeError as e:
                if from_model_name in self.__unwrapped_models:
                    logger.error(f"from model error: {from_model_name}")
                if to_model_name in self.__unwrapped_models:
                    logger.info(f"to model error: {to_model_name}")
                raise e
            if from_model_name in self.__models:
                self.__model_is_handle[from_model_name] = True
            if to_model_name in self.__models:
                self.__unwrapped_models[to_model_name].patch_reparallelization(
                    (new_layers, new_param)
                )
                self.__model_is_handle[to_model_name] = False
        elif hook == "offload":
            with cuda_tmarked("offload", CUDATimeMarkType.mem_layout):
                m = self.__unwrapped_models[hook_data["model_name"]]
                if not isinstance(m, ReaLModel):
                    raise ValueError(
                        f"Model {from_model_name} (type={type(m)}) is not a ReaLModel, "
                        f"so it can't use offload."
                    )
                m.async_offload()
        else:
            raise NotImplementedError(f"Unknown hook {hook}.")

    def __handle_all_pre_hooks(self):
        # drain request queues, handle all pending hooks, then recover the queue
        cache = []
        while True:
            try:
                request, data, handled, res = self.__request_queue.get_nowait()
                request: request_reply_stream.Payload
                if not handled:
                    while len(request.pre_hooks) > 0:
                        assert len(request.pre_hooks) == len(
                            request.pre_hook_data
                        )
                        assert not handled and res is None
                        self.__handle_one_rpc_hook(
                            request.pre_hooks.pop(0),
                            request.pre_hook_data.pop(0),
                        )
                cache.append((request, data, handled, res))
            except queue.Empty:
                break

        for c in cache:
            self.__request_queue.put_nowait(c)

    def __model_poll_step(
        self,
        request: request_reply_stream.Payload,
        data: Any,
        handled: bool,
        res: Optional[Any],
    ) -> worker_base.PollResult:
        tik = time.perf_counter()
        # self.logger.info(f"Model worker {self.__worker_index} #{request.handler}# "
        #                  f"start handle request *{request.handle_name}*, "
        #                  f"request_id {request.request_id}.")

        if isinstance(request.handler, str):
            assert request.handler == f"__data{self.__dataset_dp_rank}__"
            handler_model_name = self.__src_rpc_model_name
        else:
            handler_model_name = request.handler.model_name

        # handle post hooks
        if handled and len(request.post_hooks) > 0:
            assert len(request.post_hooks) == len(request.post_hook_data)
            self.__handle_one_rpc_hook(
                request.post_hooks.pop(0), request.post_hook_data.pop(0)
            )
            self.__request_queue.put_nowait((request, data, True, res))
            return

        if handled and len(request.post_hooks) == 0:
            self.__reply_queue.put_nowait((request, res))
            # self.logger.info(f"Model worker {self.__worker_index} #{request.handler}# "
            #                          f"finish handling request *{request.handle_name}*, "
            #                          f"request_id {request.request_id}.")
            sample_count = (
                data.length(0) if isinstance(data, namedarray.NamedArray) else 1
            )
            self.__request_sample_size[request.request_id] = sample_count
            return

        assert not handled and res is None, (
            handled,
            res,
            len(request.post_hooks),
        )

        with constants.model_scope(handler_model_name):
            res = None
            if request.handle_name == "empty":
                # Empty request is used for executing hooks, e.g., data transfer, parameter syncrhonization.
                pass
            ############## initialization ##############
            elif request.handle_name == "initialize":
                assert not self.__model_is_handle[request.handler.model_name]
                self.__models[request.handler.model_name] = (
                    self._backend.initialize(self._model, data)
                )
                self.__backend_initialized[request.handler.model_name] = True
            elif request.handle_name == "model_config":
                if isinstance(
                    self.__unwrapped_models[request.handler.model_name],
                    ReaLModel,
                ):
                    res = self.__unwrapped_models[
                        request.handler.model_name
                    ].config
            ############## data loading ##############
            elif request.handle_name == "fetch":
                fetched_data = data_api.split_sequences(self.__cur_sample)
                if self.__recover_run and not self.__recover_first_epoch_done:
                    fetched_data = list(
                        filter(
                            lambda x: hash(x)
                            not in self.__recover_info.hash_vals_to_ignore,
                            fetched_data,
                        )
                    )

                seqlens = [x.metadata["seqlens"][0] for x in fetched_data]
                res = data_api.DataBatchMeta(
                    dp_rank=self._dp_rank,
                    seqlens=seqlens,
                    keys=list(self.__cur_sample.keys()),
                    epoch=self.__dataset_epoch,
                    is_final_batch=(
                        self.__dataset_batch_counter
                        == len(self.__dataloader) - 1
                    ),
                    hash_vals=[hash(x) for x in fetched_data],
                )
                self.__fetched_sample_cache += fetched_data
                self.__cur_sample = None
            elif request.handle_name == "store":
                buffer_indices = request.data
                assert len(buffer_indices) == len(
                    self.__fetched_sample_cache
                ), (
                    len(buffer_indices),
                    len(self.__fetched_sample_cache),
                )
                for buf_idx, x in zip(
                    buffer_indices, self.__fetched_sample_cache
                ):
                    for k, v in x.items():
                        assert v.device == torch.device("cpu")
                        self.__data_owner_storage[buf_idx][k] = v
                self.__fetched_sample_cache.clear()
            elif request.handle_name == "spec":
                res = self.__dataset_n_seqs
            elif request.handle_name == "clear_data_cache":
                with cuda_tmarked("clear_data_cache", CUDATimeMarkType.misc):
                    buf_indices = request.data
                    for buf_idx in buf_indices:
                        if buf_idx in self.__data_owner_storage:
                            del self.__data_owner_storage[buf_idx]
                        if buf_idx in self.__data_receive_cache:
                            del self.__data_receive_cache[buf_idx]
                        if buf_idx in self.__data_sent_worker_indices:
                            del self.__data_sent_worker_indices[buf_idx]
                        if buf_idx in self.__data_received_worker_indices:
                            del self.__data_received_worker_indices[buf_idx]
                    if (
                        self.config.cuda_cache_cleanliness
                        and self.__clear_cache_frequency.check()
                    ):
                        st = time.monotonic()
                        gc.collect()
                        torch.cuda.empty_cache()
                        gc.collect()
                        et = time.monotonic()
                        blogger.debug(
                            f"Model worker {self.__worker_index} cleared cache in {et-st:.4f}s"
                        )
                dump_tmark_db(self.__worker_index)
            ############## computation function calls ##############
            elif request.handle_name in ["inference", "generate", "train_step"]:
                res = self.__handle_model_function_calls(request, data)
            elif request.handle_name == "evaluate":
                assert not self.__model_is_handle[
                    request.handler.model_name
                ], request.handler.model_name
                res = self._interface.evaluate(
                    self._model, self._eval_dataloader
                )  # -> Dict
            elif request.handle_name == "save":
                if not self.__model_is_handle[request.handler.model_name]:
                    self._interface.save(self._model, data)  # -> None
            else:
                raise NotImplementedError(
                    f"Unknown request type: {request.handle_name}."
                )

            if (
                request.handle_name in TIME_RECORD_RPCS
                and self._is_dp_head
                and self._dp_rank == 0
            ):
                blogger.info(
                    f"Model worker #{handler_model_name}# handle request *{request.handle_name}*"
                    f" in ${time.perf_counter() - tik:.4f}$s"
                )

        self.__request_queue.put_nowait((request, data, True, res))

    def __handle_model_function_calls(
        self, request: request_reply_stream.Payload, data: Any
    ):
        assert not self.__model_is_handle[
            request.handler.model_name
        ], request.handler.model_name
        input_queue = self.__compute_input_queues[request.handler.model_name][
            request.handle_name
        ]
        data, buffer_indices, seqlens, output_key_remap = (
            input_queue.get_nowait()
        )
        data: namedarray.NamedArray
        data.register_metadata(seqlens=seqlens)
        if request.handle_name == "inference":
            res = self._interface.inference(self._model, data)  # -> NamedArray
        elif request.handle_name == "train_step":
            res = self._interface.train_step(self._model, data)  # -> Dict
        elif request.handle_name == "generate":
            res = self._interface.generate(self._model, data)  # -> NamedArray

        if res is not None and isinstance(res, namedarray.NamedArray):
            new_res = {}
            for k, v in res.items():
                if k in output_key_remap:
                    new_res[output_key_remap[k]] = v
                else:
                    new_res[k] = v
            new_res = namedarray.from_dict(new_res)
            new_res.register_metadata(**res.metadata)
            res = new_res, buffer_indices, seqlens

        utilization = pynvml.nvmlDeviceGetUtilizationRates(self.__nvml_handle)
        memory_info = pynvml.nvmlDeviceGetMemoryInfo(self.__nvml_handle)
        total_memory = memory_info.total / (
            1024**2
        )  # Convert bytes to megabytes
        used_memory = memory_info.used / (1024**2)
        memory_usage_percentage = (used_memory / total_memory) * 100
        logger.info(
            f"Worker Index {self.__worker_index}, GPU {self.__pg_info.local_gpu_id}: "
            f"Compute Utilization - {utilization.gpu}%, "
            f"Total Memory - {total_memory:.2f}MB, Used Memory - {used_memory:.2f}MB, "
            f"Memory Usage - {memory_usage_percentage:.2f}%"
        )
        return res

    @cuda_tmark("data_transfer", CUDATimeMarkType.comm)
    def __data_transfer_among_workers(self, hook_data: Dict[str, Any]):

        comm_plan = data_transfer_comm.derive_data_transfer_plan(
            keys=hook_data["keys"],
            global_buffer_indices=hook_data["buffer_indices"],
            global_seqlens=hook_data["seqlens"],
            consumer_name=hook_data["target"],
            consumer_mapping=hook_data["target_mapping"],
            producer_names=hook_data["producer_names"],
            producer_mappings=hook_data["producer_mappings"],
            data_transfer_info=self.__data_transfer_info,
        )

        data = dict()
        for step in comm_plan:
            if (
                isinstance(step, data_transfer_comm.DataTransferReceiverStep)
                and step.rank == dist.get_rank()
            ):
                if isinstance(
                    self.__unwrapped_models[hook_data["target"]], ReaLModel
                ):
                    vocab_size = self.__unwrapped_models[
                        hook_data["target"]
                    ].config.vocab_size
                else:
                    vocab_size = None
                buf_indices = step.buf_indices
                seqlens = step.seqlens
                if step.src == dist.get_rank():
                    for buf_idx in buf_indices:
                        self.__data_owner_storage[buf_idx][step.key] = (
                            self.__data_owner_storage[buf_idx][step.key].to(
                                self.__device
                            )
                        )
                    vs = torch.cat(
                        [
                            self.__data_owner_storage[buf_idx][step.key]
                            for buf_idx in buf_indices
                        ],
                        dim=0,
                    )
                else:
                    all_sent_dst_ranks = [
                        self.__data_received_worker_indices[buf_idx][step.key]
                        for buf_idx in buf_indices
                    ]
                    if all(
                        set(step.dst_ranks).issubset(set(sent_dst_ranks))
                        for sent_dst_ranks in all_sent_dst_ranks
                    ):
                        vs = torch.cat(
                            [
                                self.__data_receive_cache[buf_idx][step.key]
                                for buf_idx in buf_indices
                            ],
                            dim=0,
                        )
                    else:
                        total_len = 0
                        for seqlen in seqlens:
                            shape = data_api.get_shape_from_key_and_seqlen(
                                step.key, seqlen, vocab_size
                            )
                            total_len += int(np.prod(shape))
                        dtype = data_api.get_dtype_from_key(step.key)
                        buf = torch.zeros(
                            (total_len,),
                            dtype=dtype,
                            device=torch.cuda.current_device(),
                        )
                        # print(f"{dist.get_rank()} recv {step.key} from {step.src} with shape {buf.shape}")
                        dist.broadcast(buf, src=step.src, group=step.group)
                        vs = buf.clone().view(-1, *shape[1:])
                        for buf_idx in buf_indices:
                            self.__data_received_worker_indices[buf_idx][
                                step.key
                            ].union(step.dst_ranks)
                offset = 0
                for seqlen, buf_idx in zip(seqlens, buf_indices):
                    shape = data_api.get_shape_from_key_and_seqlen(
                        step.key, seqlen, vocab_size
                    )
                    v = vs[offset : offset + shape[0]]
                    offset += shape[0]
                    data[(buf_idx, step.key)] = (v, seqlen)
                    if (
                        step.key not in self.__data_owner_storage[buf_idx]
                        and step.key not in self.__data_receive_cache[buf_idx]
                    ):
                        self.__data_receive_cache[buf_idx][step.key] = v

            if (
                isinstance(step, data_transfer_comm.DataTransferSenderStep)
                and step.rank == dist.get_rank()
            ):
                buf_indices = step.buf_indices
                all_sent_dst_ranks = [
                    self.__data_sent_worker_indices[buf_idx][step.key]
                    for buf_idx in buf_indices
                ]
                if all(
                    set(step.dst_ranks).issubset(set(sent_dst_ranks))
                    for sent_dst_ranks in all_sent_dst_ranks
                ):
                    pass
                else:
                    for buf_idx in buf_indices:
                        self.__data_owner_storage[buf_idx][step.key] = (
                            self.__data_owner_storage[buf_idx][step.key].to(
                                self.__device
                            )
                        )
                    vs = torch.cat(
                        [
                            self.__data_owner_storage[buf_idx][step.key]
                            for buf_idx in buf_indices
                        ],
                        dim=0,
                    )
                    if vs.dtype != data_api.get_dtype_from_key(step.key):
                        raise ValueError(
                            f"Infered dtype of {step.key} ({data_api.get_dtype_from_key(step.key)})"
                            f" is not equal to the actual dtype ({vs.dtype}). "
                            "Is it correctly set in the dataset implementation?"
                        )
                    # print(f"{dist.get_rank()} send {step.key} to {step.dst_ranks} with shape {vs.shape}")
                    dist.broadcast(vs, src=step.rank, group=step.group)
                    for buf_idx in buf_indices:
                        self.__data_sent_worker_indices[buf_idx][
                            step.key
                        ].union(step.dst_ranks)

        if len(data) > 0:
            local_buffer_indices = sorted(
                list(set([buf_idx for buf_idx, _ in data.keys()]))
            )
            local_keys = list(set([key for _, key in data.keys()]))

            local_seqlens = []
            for buf_idx in local_buffer_indices:
                local_seqlens.append(data[(buf_idx, local_keys[0])][1])

            input_key_remap = hook_data["input_key_remap"]
            _data = []
            for buf_idx in local_buffer_indices:
                d = {}
                for k in local_keys:
                    v, seqlen = data[(buf_idx, k)]
                    if k in input_key_remap:
                        k = input_key_remap[k]
                    d[k] = v
                x = namedarray.from_dict(d)
                x.register_metadata(seqlens=[seqlen])
                _data.append(x)
            r = data_api.gather_sequences(_data)
            self.__compute_input_queues[hook_data["target"]][
                hook_data["handle_name"]
            ].put_nowait(
                (
                    r,
                    local_buffer_indices,
                    local_seqlens,
                    hook_data["output_key_remap"],
                )
            )

    def __post_one_response(self, request: request_reply_stream.Payload, res):
        if isinstance(res, tuple) and isinstance(res[0], namedarray.NamedArray):
            res, buffer_indices, seqlens = res
            reply = request_reply_stream.Payload(
                handler="master",
                request_id=request.request_id,
                handle_name=request.handle_name,
                data={
                    "keys": list(res.keys()),
                    "seqlens": res.metadata["seqlens"],
                    "buffer_indices": list(buffer_indices),
                },
            )
        else:
            reply = request_reply_stream.Payload(
                handler="master",
                request_id=request.request_id,
                handle_name=request.handle_name,
                data=res,
            )

        self.__stream.post(reply)
        # logger.info(f"handle_name {request.handle_name} Posted req id = {request.request_id}")

        if isinstance(request.handler, system_api.ModelShardID) and isinstance(
            res, namedarray.NamedArray
        ):
            with constants.model_scope(request.handler.model_name):
                if self._is_dp_head:
                    xs = data_api.split_sequences(res)
                    for buffer_idx, x in zip(buffer_indices, xs):
                        for k, v in x.items():
                            self.__data_owner_storage[buffer_idx][k] = v

    @cuda_tmark("post_response", CUDATimeMarkType.misc)
    def __maybe_post_responses(self):
        ready_to_post = []
        try:
            request, res = self.__reply_queue.get_nowait()
            ready_to_post.append((request, res))
        except queue.Empty:
            pass

        batch_size = sample_size = 0
        for request, res in ready_to_post:
            request: request_reply_stream.Payload
            self.__post_one_response(request, res)
            sample_size += self.__request_sample_size.pop(request.request_id)
            batch_size += 1
        return worker_base.PollResult(
            sample_count=sample_size, batch_count=batch_size
        )

    def __maybe_receive_one_request(self):
        try:
            r: request_reply_stream.Payload = self.__stream.poll()
            if r.handle_name == "ack":
                self.__ack_cache[r.request_id] = r
            else:
                self.__stream.post(
                    request_reply_stream.Payload(
                        handler="master",
                        request_id=r.syn_reply_id,
                        handle_name="syn",
                    ),
                )
                self.__request_cache.append(r)
        except request_reply_stream.NoMessage:
            return

    @cuda_tmark("receive_request", CUDATimeMarkType.misc)
    def __maybe_receive_requests(self):
        for _ in range(8):
            self.__maybe_receive_one_request()

        while len(self.__request_cache) > 0:
            request: request_reply_stream.Payload = self.__request_cache[0]
            while request.ack_reply_id not in self.__ack_cache:
                self.__maybe_receive_one_request()

            self.__ack_cache.pop(request.ack_reply_id)
            self.__request_cache.pop(0)

            self.__request_queue.put_nowait(
                (request, request.data, False, None)
            )

    def _poll(self):
        if not self.__dist_env_resolved:
            self.__lazy_setup()
            pynvml.nvmlInit()
            self.__nvml_handle = pynvml.nvmlDeviceGetHandleByIndex(
                self.__pg_info.local_gpu_id
            )
            self.__dist_env_resolved = True

        if self.__has_dataset:
            self.__prefetch_from_dataset()

        st = time.monotonic()
        self.__maybe_receive_requests()

        # NOTE: We ensure that all model workers have the same set of requests
        # at any time through a TCP-like protocol, i.e., req -> ack -> syn -> resp.
        # Each request is composed of pre-hooks, the main request, and post-hooks.
        # We execute all pre-hooks first because they involve data transfer
        # among workers. Then, we round-robinly execute hooks and requests.
        # These are designed to prevent mutual blocking when different requests
        # are handled in different but intersected sets of model workers.
        # E.g., If we have a request A and a request B, the execution order will be
        # A.pre_hook -> B.pre_hook -> A -> B -> A.post_hook -> B.post_hook.
        self.__handle_all_pre_hooks()
        for _ in range(16):
            try:
                request, data, handled, res = self.__request_queue.get_nowait()
                self.__model_poll_step(request, data, handled, res)
            except queue.Empty:
                break

        r = self.__maybe_post_responses()

        t = time.monotonic() - st
        self.__total_time += t
        return r

    def __recover_save(self):
        # store model and dataset states for recover
        if self.__dist_env_resolved:

            for model_name, model in self.__models.items():
                if self.__model_is_handle[model_name]:
                    continue
                constants._model_name = None  # force quit model_scope
                with constants.model_scope(model_name):
                    ckpt_save_dir = os.path.join(
                        self.__recover_states_root, "ckpt", model_name.role
                    )
                    # replace old recover ckpt
                    logger.info(
                        f"saving model {model_name} ckpt for recover at {ckpt_save_dir}. "
                        f"epoch {model.version.epoch}, epoch_step {model.version.epoch_step}, "
                        f"global step {model.version.global_step}"
                    )
                    if self.__has_dataset:
                        logger.info(
                            f"Dataset info: "
                            f"dataset epoch {self.__dataset_epoch}"
                        )
                    self._interface.save(model, ckpt_save_dir)
                    logger.info(f"saving done.")

    def _exit_hook(self, exit_status: worker_base.WorkerServerStatus):
        logger.info(
            f"Model worker {self.__worker_index} exit with status {exit_status}."
        )
        if os.environ.get("SAVE_RECOVER_STATES", "0") == "0":
            return
        if exit_status == worker_base.WorkerServerStatus.ERROR:
            try:
                sleep_time = 600
                current_sleep_time = 0
                while current_sleep_time < sleep_time:
                    logger.info(
                        f"ERROR exit, waited {current_sleep_time} s for interruption ..."
                    )
                    time.sleep(10)
                    current_sleep_time += 10
            except KeyboardInterrupt:
                logger.info("Received SIGINT, starting recover save")

        self.__recover_save()



# Estimate a fucntion-call level execution time for device mesh enumerate pruning
# assume one batch of data passes through all rpcs once
from collections import defaultdict
from typing import Optional
import argparse
import getpass
import itertools
import os
import pickle

import numpy as np
import pandas as pd

from realhf.api.core.dfg import MFCDef, ModelFamily, ModelInterfaceType
from realhf.api.core.model_api import ReaLModelConfig
from realhf.api.quickstart.model import ParallelismConfig
from realhf.search_engine.param_realloc import estimate_param_realloc_time_cost
from realhf.search_engine.utils import load_model_config
import realhf.base.cluster
import realhf.base.constants as constants
import realhf.base.logging as logging

logger = logging.getLogger("estimate", "benchmark")

PROFILE_RESULT_PATH = os.path.join(
    realhf.base.cluster.spec.fileroot,
    "logs",
    getpass.getuser(),
    "profile",
    "profile",
    "layer_stats",
)


def get_param_realloc_stats(
    model_family: ModelFamily,
    model_path: str,
    n_nodes: int,
    use_cache: bool = True,
):
    non_critic = ModelFamily(model_family._class, model_family.size, False)
    table_path = os.path.join(
        constants.PROFILER_CACHE_PATH,
        "param_realloc",
        f"prtc_{non_critic}_n{n_nodes}.pkl",
    )
    if not os.path.exists(table_path):
        print(
            f"Calculating estimation of param realloc time cost for {model_family} at {model_path}"
        )
        estimate_param_realloc_time_cost(n_nodes, {non_critic: model_path})

    print(f"Loading param realloc stats from {table_path}")
    return pickle.load(open(table_path, "rb"))


def get_organized_op_stats(
    model_family: ModelFamily, model_path: str, use_cache: bool = True
):
    non_critic = ModelFamily(model_family._class, model_family.size, False)
    # parse raw stats into list of OpInfo used for estimation
    cache_path = os.path.join(
        constants.PROFILER_CACHE_PATH, "organized_stats", f"{non_critic}.pkl"
    )
    if use_cache and os.path.exists(cache_path):
        with open(cache_path, "rb") as f:
            return pickle.load(f)

    raw_result_path = os.path.join(
        constants.PROFILER_CACHE_PATH, "layer_stats", str(non_critic)
    )
    if not os.path.exists(raw_result_path):
        from realhf.apps.main import _main_profile_layers

        _main_profile_layers(non_critic, model_path)

    raw_stats_list = []
    for fn in os.listdir(raw_result_path):
        if not fn.startswith("layer-stats"):
            continue
        num_mp = int(fn.replace(".pkl", "").split("_")[1])
        rank = int(fn.replace(".pkl", "").split("_")[2])
        with open(os.path.join(raw_result_path, fn), "rb") as f:
            stats = pickle.load(f)

        if isinstance(stats, dict):
            stats = pd.DataFrame(stats)
        elif isinstance(stats, pd.DataFrame):
            pass
        else:
            raise ValueError(f"Unsupported stats type {type(stats)}")

        stats["num_mp"] = num_mp
        stats["rank"] = rank
        raw_stats_list.append(stats)

    raw_stats = pd.concat(raw_stats_list)
    bs_list = raw_stats["bs"].unique()
    seq_len_list = raw_stats["seq_len"].unique()
    op_name_list = raw_stats["op_name"].unique()
    layer_name_list = raw_stats["layer_name"].unique()
    num_mp_list = raw_stats["num_mp"].unique()
    organized_stats = defaultdict(list)

    for op_name, bs, seq_len, layer_name, num_mp in itertools.product(
        op_name_list, bs_list, seq_len_list, layer_name_list, num_mp_list
    ):
        filter_cond = (
            (raw_stats["op_name"] == op_name)
            & (raw_stats["bs"] == bs)
            & (raw_stats["seq_len"] == seq_len)
            & (raw_stats["layer_name"] == layer_name)
            & (raw_stats["num_mp"] == num_mp)
        )
        avg_time_ns = raw_stats[filter_cond]["time_ns"].mean()
        x = int(bs) if op_name == "fwd_gen_1" else int(bs * seq_len)

        organized_stats["op_name"].append(op_name)
        organized_stats["layer_name"].append(layer_name)
        organized_stats["bs"].append(bs)
        organized_stats["seq_len"].append(seq_len)
        organized_stats["num_mp"].append(num_mp)
        organized_stats["avg_time_ns"].append(avg_time_ns)
        organized_stats["x"].append(x)

    df = pd.DataFrame(organized_stats)
    if use_cache:
        os.makedirs(os.path.dirname(cache_path), exist_ok=True)
        with open(cache_path, "wb") as f:
            pickle.dump(df, f)

    return df


def computation_instruction_time_cost(
    op_stats: pd.DataFrame,
    op_name: str,
    num_layers: int,
    parallel_strategy: ParallelismConfig,
    bs: int,
    seqlen: int,
):
    # inst cost unit: ns
    layer_names = ["embedding_layer", "block_0", "head"]
    num_pp = parallel_strategy.pipeline_parallel_size
    num_mp = parallel_strategy.model_parallel_size
    op_stats = op_stats[
        (op_stats["op_name"] == op_name) & (op_stats["num_mp"] == num_mp)
    ]

    op_cost = {}
    embed_stats = op_stats[op_stats["layer_name"] == "embedding_layer"]
    if embed_stats[
        (embed_stats["bs"] == bs) & (embed_stats["seq_len"] == seqlen)
    ].empty:
        # do linear interpolation for data points that does not exist
        for layer_name in layer_names:
            layer_stats = op_stats[op_stats["layer_name"] == layer_name]
            assert layer_stats[
                (layer_stats["bs"] == bs) & (layer_stats["seq_len"] == seqlen)
            ].empty
            assert not layer_stats.empty, (
                layer_name,
                op_name,
                num_mp,
                op_stats,
            )
            xs = layer_stats["x"]
            ys = layer_stats["avg_time_ns"]
            x = int(bs) if op_name == "fwd_gen_1" else int(bs * seqlen)
            y = np.interp(x, xs, ys)
            if max(xs) < x or min(xs) > x:
                logger.warning(
                    f"Interpolated value outside profiling range, "
                    f"parallel strategy {parallel_strategy}: "
                    f"{x} in {sorted(list(set(xs)))}"
                )
                # estimate using largest or smallest value
                if max(xs) < x:
                    y = ys.max() * (x / xs[ys.idxmax()])
                else:
                    y = ys.min() * (x / xs[ys.idxmin()])
            op_cost[layer_name] = y
    else:
        for layer_name in layer_names:
            assert not op_stats[op_stats["layer_name"] == layer_name].empty
            required_stats = op_stats[
                (op_stats["layer_name"] == layer_name)
                & (op_stats["bs"] == bs)
                & (op_stats["seq_len"] == seqlen)
            ]
            assert required_stats.shape[0] == 1
            op_cost[layer_name] = required_stats["avg_time_ns"].values[0]

    embedding_layer_cost = op_cost["embedding_layer"]
    block_0_cost = op_cost["block_0"]
    head_cost = op_cost["head"]
    cost = (
        embedding_layer_cost + num_layers * block_0_cost + head_cost
    ) / num_pp
    return cost


def communication_instruction_time_cost(comm_stats, size, comm_type):
    return size / comm_stats[comm_type]  # unit: ns


def estimate_instruction_time_costs(
    model_family: ModelFamily,
    model_path: str,
    num_layers: int,  # model configuration, num transformers layer
    parallel_strategy: ParallelismConfig,
    hidden_dim: int,
    batch_size: int,
    seq_len: int,
    n_ppo_minibatches: int = 1,
):
    comm_stats = default_communication_stats()
    op_stats = get_organized_op_stats(model_family, model_path, use_cache=True)

    num_mp = parallel_strategy.model_parallel_size
    num_pp = parallel_strategy.pipeline_parallel_size
    num_dp = parallel_strategy.data_parallel_size
    num_gpus = num_dp * num_mp * num_pp

    train_mbs = (
        batch_size / (2 * num_pp * num_dp * n_ppo_minibatches)
        if num_pp > 1
        else batch_size / (num_dp * n_ppo_minibatches)
    )
    gen_mbs = batch_size / (num_pp * num_dp)

    # pprint.pprint(op_cost, indent=4)
    inst_keys = [
        "gen_fwd_0",
        "gen_fwd_1",
        "inf_fwd",
        "train_fwd",
        "train_bwd",
        "train_opt",
    ]
    op_names = ["fwd_gen_0", "fwd_gen_1", "fwd", "fwd", "bwd", "opt"]
    inst_stats = {}

    for inst_key, op_name in zip(inst_keys, op_names):
        mbs = train_mbs if "train" in inst_key else gen_mbs
        inst_stats[inst_key] = computation_instruction_time_cost(
            op_stats, op_name, num_layers, parallel_strategy, mbs, seq_len
        )

    comm_type = "remote_send" if num_gpus // num_pp >= 8 else "local_send"

    inst_stats["act_p2p"] = communication_instruction_time_cost(
        comm_stats, 2 * hidden_dim * train_mbs * seq_len, comm_type
    )
    inst_stats["grad_p2p"] = communication_instruction_time_cost(
        comm_stats, 2 * hidden_dim * train_mbs * seq_len, comm_type
    )
    inst_stats["gen_act_p2p"] = communication_instruction_time_cost(
        comm_stats, 2 * hidden_dim * gen_mbs, comm_type
    )
    return inst_stats


def _estimate_rpc_time_cost(
    inst_stats,
    parallel_strategy: ParallelismConfig,
    model_interface_type: ModelInterfaceType,
    # model function call args
    num_gen_tokens: int,
    gradient_checkpointing: bool,
    n_ppo_minibatches: int = 1,
):
    # TODO: improve/remove heuristic
    num_pp = parallel_strategy.pipeline_parallel_size
    num_dp = parallel_strategy.data_parallel_size
    num_mp = parallel_strategy.model_parallel_size
    if model_interface_type == ModelInterfaceType.INFERENCE:
        num_micro_batches = num_pp
        compute_cost = inst_stats["inf_fwd"] * (num_pp + num_micro_batches - 1)
        comm_cost = inst_stats["act_p2p"] * (num_pp + num_micro_batches - 2) * 2
        if num_mp > 1:  # and num_pp * num_dp > 1:
            compute_cost = compute_cost * (1 - num_mp * 0.03)
    elif model_interface_type == ModelInterfaceType.TRAIN_STEP:
        # TODO: add reduce grads, add ppo micro batches
        num_micro_batches = num_pp * 2 if num_pp > 1 else 1
        compute_cost = (inst_stats["train_fwd"] + inst_stats["train_bwd"]) * (
            num_pp + num_micro_batches - 1
        ) + inst_stats["train_opt"]
        if gradient_checkpointing:
            compute_cost += inst_stats["train_fwd"] * (
                num_pp + num_micro_batches - 1
            )
        comm_cost = (
            (inst_stats["grad_p2p"] + inst_stats["act_p2p"])
            * (num_pp + num_micro_batches - 2)
            * 2
        )
        compute_cost = compute_cost * n_ppo_minibatches
        comm_cost = comm_cost * n_ppo_minibatches
        if num_pp * num_dp <= 1:
            compute_cost = compute_cost * (1 - num_mp * 0.04)
        if num_mp > 1:  # and num_pp * num_dp > 1:
            compute_cost = compute_cost * (1 - num_mp * 0.03)
    elif model_interface_type == ModelInterfaceType.GENERATE:
        num_micro_batches = num_pp
        num_gen_tokens = num_gen_tokens
        compute_cost = (
            inst_stats["gen_fwd_0"] * (num_pp + num_micro_batches - 1)
            + inst_stats["gen_fwd_1"] * (num_gen_tokens - 1) * num_micro_batches
        )

        if num_dp * num_mp > 1:
            compute_cost = compute_cost * (1 - min(num_dp * num_mp, 8) * 0.03)
        comm_cost = 0

    # dirty heuristic
    if num_pp > 8:
        compute_cost = compute_cost * (1 + num_pp * 0.01)

    # FIXME: disable comm cost for its not accurate
    comm_cost = 0

    return compute_cost + comm_cost


def estimate_rpc_time_cost(
    rpc: MFCDef,
    parallel_strategy: ParallelismConfig,
    bs: int,
    seq_len: int,
    num_gen_tokens: int = 256,
    gradient_checkpointing: bool = False,
    n_ppo_minibatches: int = 1,
):
    # time unit: miliseconds
    # FIXME: n_ppo_minibatches > 1 will result in bad estimation
    # when batch size is large enough, n_ppo_minibatches > 1 will not affect the estimation
    n_ppo_minibatches = 1
    model_type = rpc.model_type
    model_path = rpc.model_path
    model_config = load_model_config(rpc.model_type._class, rpc.model_path)

    inst_cost = estimate_instruction_time_costs(
        model_type,
        model_path,
        model_config.n_layers,
        parallel_strategy,
        model_config.hidden_dim,
        bs,
        seq_len,
        n_ppo_minibatches=n_ppo_minibatches,
    )
    return (
        _estimate_rpc_time_cost(
            inst_cost,
            parallel_strategy,
            rpc.interface_type,
            num_gen_tokens=num_gen_tokens,
            gradient_checkpointing=gradient_checkpointing,
            n_ppo_minibatches=n_ppo_minibatches,
        )
    ) / 1e6


def default_communication_stats(if_print=False):
    # use default comm stats of cluster
    r = dict(
        # between GPUs on the same node
        local_send=170,  # unit: GB/s
        local_recv=170,
        # IB between GPUs on different nodes
        remote_send=20,  # unit: GB/s
        remote_recv=20,
    )
    if if_print:
        print(r)
    return r


def estimate_model_size(model_config: ReaLModelConfig):
    h = model_config.hidden_dim
    i = model_config.intermediate_dim
    v = model_config.vocab_size
    L = model_config.n_layers
    # for llama actor only
    n_params = 3 * v * h + (3 * h * i + 4 * h * h) * L
    return 2 * n_params


def estimate_rpc_memory_cost(
    rpc: MFCDef,
    parallel_strategy: ParallelismConfig,
    batch_size: int,
    seq_len: int,
    offload: bool = False,
    gradient_checkpointing: bool = False,
    offload_optimizer: bool = False,
    n_ppo_minibatches: int = 1,
    num_gen_tokens: int = 128,
):
    # TODO: improve heuristic
    interface_type = rpc.interface_type
    model_config = load_model_config(rpc.model_type._class, rpc.model_path)

    h = model_config.hidden_dim
    i = model_config.intermediate_dim
    v = model_config.vocab_size
    s = seq_len
    gs = num_gen_tokens
    b = batch_size
    L = model_config.n_layers
    # for llama actor only
    n_params = 3 * v * h + (3 * h * i + 4 * h * h) * L
    param_mem = 2 * n_params
    grad_mem = 2 * n_params
    optimizer_mem = 20 * n_params if not offload_optimizer else 0

    num_pp = parallel_strategy.pipeline_parallel_size
    num_mp = parallel_strategy.model_parallel_size
    num_dp = parallel_strategy.data_parallel_size
    # zero1, pp and mp divide evenly
    # enable sequence parallel
    if interface_type == ModelInterfaceType.TRAIN_STEP:
        # gradient checkpointing is always enabled for flash attn
        static_mem = (param_mem + grad_mem) // (
            num_pp * num_mp
        ) + optimizer_mem // (num_pp * num_dp * num_mp)
        micro_bs = b // (2 * num_pp * num_dp) if num_pp > 0 else b // (num_dp)
        if gradient_checkpointing:
            active_mem = (micro_bs * s * h * num_pp * 2) // (num_pp * num_mp)
        else:
            # FIXME: calculate other memory entries
            active_mem = (
                (micro_bs * s * h * num_pp * 2) * 2 * L // (num_pp * num_mp)
            )
        return static_mem + active_mem, static_mem
    elif interface_type == ModelInterfaceType.INFERENCE:
        static_mem = int(2 * param_mem // (num_pp * num_mp))
        # if num_dp > 4:
        #     static_mem = static_mem * 1.25
        if offload:
            return static_mem, 0  # assume offload
        else:
            return static_mem, static_mem
    elif interface_type == ModelInterfaceType.GENERATE:
        static_mem = int(2 * param_mem // (num_pp * num_mp))
        if num_dp > 4 and num_dp * num_mp * num_pp <= 16:
            static_mem = static_mem * 1.25
        if num_mp == 0 and num_pp == 0:
            static_mem = static_mem * 1.25
        active_mem = (
            2 * (2 * b * (gs + s) * h) * L // (num_pp * num_mp * num_dp)
        )  # kv cache
        return static_mem + active_mem, static_mem


def example(rpcs):
    if args.model_size == 7:
        n_nodes = 1
    elif args.model_size == 13:
        n_nodes = 2
    elif args.model_size == 34:
        n_nodes = 4
    elif args.model_size == 70:
        n_nodes = 8

    expr_name = f"profile-s{args.model_size}p{n_nodes}m1d8"

    rollout, inf, train = rpcs

    bs = 128
    seq_len = 1024

    p1 = ParallelismConfig(
        pipeline_parallel_size=1, model_parallel_size=4, data_parallel_size=8
    )
    rpc_cost = estimate_rpc_time_cost(
        train,
        p1,
        gradient_checkpointing=True,
        num_gen_tokens=896,
        bs=bs,
        seq_len=seq_len,
        n_ppo_minibatches=4,
    )
    mem_cost, static_mem = estimate_rpc_memory_cost(rollout, p1, bs, seq_len)
    print(
        f"{p1} rpc cost {rpc_cost:.2f} seconds mem cost {mem_cost/(1024**3):.2f} GB"
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run a profiling experiment.")
    parser.add_argument(
        "-s",
        "--model_size",
        type=int,
        default=7,
    )
    args = parser.parse_args()

    example(args)



from typing import *
import dataclasses

import realhf.base.topology as topology


@dataclasses.dataclass
class Dataset:
    type_: str
    args: Dict[str, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class DataLoader:
    type_: str = "default"
    args: Dict[str, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class ModelWrapper:
    type_: str
    args: Dict[str, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class Model:
    type_: str
    args: Dict[str, Any] = dataclasses.field(default_factory=dict)
    wrappers: List[ModelWrapper] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class ModelBackend:
    type_: str
    args: Dict[str, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass(unsafe_hash=True, order=True)
class ModelName:
    """A unique identifier for a model.

    :param role: The role of the model, e.g. "actor", "critic".
    :type role: str
    :param replica_id: The replica id of the model.
    :type replica_id: int
    """

    role: str
    replica_id: int

    @property
    def name(self):
        return str(self)


@dataclasses.dataclass(unsafe_hash=True)
class ModelFamily:
    _class: str
    size: int
    is_critic: bool

    def __repr__(self):
        s = f"{self._class}-{self.size}"
        if self.is_critic:
            s += "-critic"
        return s


@dataclasses.dataclass
class ModelShardID:
    model_name: ModelName
    dp_rank: int
    mp_rank: int
    pp_rank: int
    topo: topology.PipeModelDataParallelTopology = dataclasses.field(
        default_factory=lambda: topology.PipeModelDataParallelTopology(1, 1, 1)
    )

    def __post_init__(self):
        assert self.dp_rank >= 0 and self.mp_rank >= 0 and self.pp_rank >= 0
        if "@" in self.model_name.role:
            raise ValueError("model_name cannot contain @")
        assert self.dp_rank < self.topo.get_dim("data")
        assert self.mp_rank < self.topo.get_dim("model")
        assert self.pp_rank < self.topo.get_dim("pipe")

    @property
    def parallelism_rank(self):
        return self.topo.get_rank(
            data=self.dp_rank, model=self.mp_rank, pipe=self.pp_rank
        )

    @classmethod
    def from_parallelism_rank(cls, model_name, topo, parallelism_rank):
        c = topo.get_coord(parallelism_rank)
        return cls(
            model_name=model_name,
            dp_rank=c.data,
            mp_rank=c.model,
            pp_rank=c.pipe,
            topo=topo,
        )

    def __repr__(self):
        return f"{self.model_name}@pp{self.pp_rank:02d}@mp{self.mp_rank:02d}@dp{self.dp_rank:02d}"

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        # Compare the key attribute for equality
        if isinstance(other, ModelShardID):
            return (
                self.model_name == other.model_name
                and self.dp_rank == other.dp_rank
                and self.mp_rank == other.mp_rank
                and self.pp_rank == other.pp_rank
            )
        return False


@dataclasses.dataclass
class StandaloneModelShard:
    id: ModelShardID
    model: Model
    backend: ModelBackend
    # evaluation
    eval_datasets: Optional[List[Dataset]] = None
    eval_dataloader: Optional[DataLoader] = DataLoader(
        "packed_eval", args=dict(batch_size=128)
    )
    should_instantiate: bool = True



from typing import Any, Dict, List, Optional, Tuple
import dataclasses
import enum
import getpass
import os
import queue
import re
import socket
import threading
import time

from realhf.base import (
    cluster,
    logging,
    monitor,
    name_resolve,
    names,
    network,
    timeutil,
)
from realhf.base.gpu_utils import set_cuda_device
import realhf.api.core.system_api as system_api

logger = logging.getLogger("worker")

_MAX_SOCKET_CONCURRENCY = 1000
WORKER_WAIT_FOR_CONTROLLER_SECONDS = 3600
WORKER_JOB_STATUS_LINGER_SECONDS = 60


class WorkerException(Exception):

    def __init__(self, worker_name, worker_status, scenario):
        super(WorkerException, self).__init__(
            f"Worker {worker_name} is {worker_status} while {scenario}"
        )
        self.worker_name = worker_name
        self.worker_status = worker_status
        self.scenario = scenario


class WorkerServerStatus(str, enum.Enum):
    """List of all possible Server status. This is typically set by workers hosting the server, and
    read by the controller.
    """

    READY = "READY"
    RUNNING = "RUNNING"
    PAUSED = "PAUSED"
    COMPLETED = "COMPLETED"

    UNKNOWN = "UNKNOWN"  # CANNOT be set.
    INTERRUPTED = "INTERRUPTED"
    ERROR = "ERROR"
    LOST = "LOST"  # CANNOT be set


class NoRequstForWorker(Exception):
    pass


class WorkerServerTaskQueue:

    def try_get_request(self) -> Tuple[str, Dict[str, Any]]:
        raise NotImplementedError()

    def respond(self, response):
        raise NotImplementedError()

    @property
    def port(self) -> int:
        return -1


class WorkerServer:
    """A light-weight implementation of an RPC server.

    Note that this server only allows a single client connection for now, as that is sufficient for
    workers which respond to the controller only.

    Example:
        # Server side.
        server = RpcServer(port)
        server.register_handler('foo', foo)
        while True:
            server.handle_requests()

        # Client side.
        client = RpcClient(host, port)
        client.request('foo', x=42, y='str') # foo(x=42, y='str') will be called on the server side.
    """

    def __init__(
        self,
        worker_name,
        experiment_name,
        trial_name,
        task_queue: WorkerServerTaskQueue,
    ):
        """Specifies the name of the worker that WorkerControlPanel can used to find and manage.
        Args:
            worker_name: Typically "<worker_type>/<worker_index>".
        """
        self.__worker_name = worker_name
        self.__experiment_name = experiment_name
        self.__trial_name = trial_name

        self.__task_queue = task_queue

        self.__handlers = {}
        host_ip = socket.gethostbyname(socket.gethostname())

        try:
            controller_status = name_resolve.wait(
                names.worker_status(experiment_name, trial_name, "ctl"),
                timeout=WORKER_WAIT_FOR_CONTROLLER_SECONDS,
            )
        except TimeoutError:
            raise TimeoutError(
                f"Worker ({experiment_name, trial_name, worker_name}) connect to controller timeout from host {socket.gethostname()}."
            )

        if controller_status != "READY":
            raise RuntimeError(
                f"Abnormal controller state on experiment launch {controller_status}."
            )

        if experiment_name is not None and trial_name is not None:
            key = names.worker(experiment_name, trial_name, worker_name)
            address = f"{host_ip}:{self.__task_queue.port}"
            name_resolve.add(
                key, address, keepalive_ttl=10, delete_on_exit=True
            )
            logger.debug(
                "Added name_resolve entry %s for worker server at %s",
                key,
                address,
            )

    def register_handler(self, command, fn):
        """Registers an RPC command. The handler `fn` shall be called when `self.handle_requests()` sees an
        incoming command of the registered type.
        """
        if command in self.__handlers:
            raise KeyError(f"Command '{command}' exists")
        self.__handlers[command] = fn

    def handle_requests(self, max_count=None):
        """Handles queued requests in order, optionally limited by `max_count`.

        Returns:
            The count of requests handled.
        """
        count = 0
        while max_count is None or count < max_count:
            try:
                command, kwargs = self.__task_queue.try_get_request()
            except NoRequstForWorker:
                # Currently no request in the queue.
                break
            logger.debug("Handle request %s with kwargs %s", command, kwargs)
            if command in self.__handlers:
                try:
                    response = self.__handlers[command](**kwargs)
                    logger.debug("Handle request: %s, ok", command)
                except WorkerException:
                    raise
                except Exception as e:
                    logger.error("Handle request: %s, error", command)
                    logger.error(e, exc_info=True)
                    response = e
            else:
                logger.error("Handle request: %s, no such command", command)
                response = KeyError(f"No such command: {command}")
            self.__task_queue.respond(response)
            logger.debug("Handle request: %s, sent reply", command)
            count += 1
        return count

    def set_status(self, status: WorkerServerStatus):
        """On graceful exit, worker status is cleared."""
        name_resolve.add(
            names.worker_status(
                experiment_name=self.__experiment_name,
                trial_name=self.__trial_name,
                worker_name=self.__worker_name,
            ),
            value=status.value,
            keepalive_ttl=WORKER_JOB_STATUS_LINGER_SECONDS,  # Job Status lives one minutes after worker exit.
            replace=True,
            delete_on_exit=False,
        )


class WorkerControlPanelRequester:

    class Future:

        def result(self, timeout=None):
            raise NotImplementedError()

    def async_request(
        self,
        worker_name: str,
        address: str,
        command: str,
        wait_for_response: bool = True,
        **kwargs,
    ) -> Future:
        raise NotImplementedError()


class WorkerControlPanel:
    """A class that defines the management utilities to all the workers of an experiment trial."""

    @dataclasses.dataclass
    class Response:
        worker_name: str
        result: WorkerControlPanelRequester.Future
        timed_out: bool = False

    def __init__(
        self,
        experiment_name,
        trial_name,
        requester: WorkerControlPanelRequester,
    ):
        self.__closed = False

        self.__experiment_name = experiment_name
        self.__trial_name = trial_name
        self.__worker_addresses = {}

        self.__requester = requester

        self.__logger = logging.getLogger("worker control panel")

    def __del__(self):
        if not self.__closed:
            self.close()
            self.__closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.__closed:
            self.close()
            self.__closed = True

    def close(self):
        self.__logger.info("Closing worker control panel.")

    @staticmethod
    def name(worker_type, worker_index):
        return f"{worker_type}/{worker_index}"

    @staticmethod
    def parse_name(worker_name):
        type_, index = worker_name.split("/")
        return type_, index

    @property
    def worker_names(self) -> List[str]:
        """Returns current connected workers. A WorkerControlPanel initializes with no connected workers.
        Workers are connected via either self.connect(names), or self.auto_connect().
        """
        return list(self.__worker_addresses.keys())

    def connect(
        self,
        worker_names: List[str],
        timeout=None,
        raises_timeout_error=False,
        reconnect=False,
        progress=False,
    ) -> List[str]:
        """Waits until all the workers specified by the given names are ready for receiving commands.

        Args:
            worker_names: A list of workers to connect.
            timeout: The maximum waiting time in seconds, or None if infinite.
            raises_timeout_error: If True, any connection failure will result in the raise of an TimeoutError.
                If False, such exception can be detected via the returned succeeded list.
            reconnect: If True, this will reconnect to the workers that has already connected. If False, any
                worker in `worker_names` that has already connected will be ignored.
            progress: Whether to show a progress bar.

        Returns:
            A list of successfully connected or reconnected workers. If a specified worker is missing, it can
            either be that it is already connected and `reconnect` is False, or that the connection timed out.
        """
        rs = []
        deadline = time.monotonic() + (timeout or 0)
        if progress:
            try:
                import tqdm

                worker_names = tqdm.tqdm(worker_names, leave=False)
            except ModuleNotFoundError:
                pass
        for name in worker_names:
            if name in self.__worker_addresses:
                if reconnect:
                    del self.__worker_addresses[name]
                else:
                    continue
            try:
                if timeout is not None:
                    timeout = max(0, deadline - time.monotonic())
                self.__logger.debug(
                    f"Connecting to worker {name}, timeout {timeout}"
                )
                server_address = name_resolve.wait(
                    names.worker(
                        self.__experiment_name, self.__trial_name, name
                    ),
                    timeout=timeout,
                )
                self.__logger.debug(f"Connecting to worker {name} done")
            except TimeoutError as e:
                if raises_timeout_error:
                    raise e
                continue
            # self.__worker_addresses[name] stores address
            self.__worker_addresses[name] = server_address
            rs.append(name)
        return rs

    def auto_connect(self) -> List[str]:
        """Auto-detects available workers belonging to the experiment trial, and connects to them.

        Returns:
            Names of successfully connected workers.
        """
        name_root = names.worker_root(self.__experiment_name, self.__trial_name)
        worker_names = [
            r[len(name_root) :] for r in name_resolve.find_subtree(name_root)
        ]
        return self.connect(worker_names, timeout=0, raises_timeout_error=True)

    def request(self, worker_name: str, command, **kwargs) -> Any:
        """Sends an request to the specified worker."""
        address = self.__worker_addresses[worker_name]
        return self.__requester.async_request(
            worker_name, address, command, **kwargs
        ).result()

    def group_request(
        self,
        command,
        worker_names: Optional[List[str]] = None,
        worker_regex: Optional[str] = None,
        timeout=None,
        progress=False,
        worker_kwargs: Optional[List[Dict[str, Any]]] = None,
        wait_response=True,
        **kwargs,
    ) -> List[Response]:
        """Requests selected workers, or all connected workers if not specified.

        Args:
            command: RPC command.
            worker_names: Optional selection of workers.
            worker_regex: Optional regex selector of workers.
            timeout: Optional timeout.
            progress: Whether to show a progress bar.
            worker_kwargs: RPC arguments, but one for each worker instead of `kwargs` where every worker share
                the arguments. If this is specified, worker_names must be specified, and worker_regex, kwargs
                must be None.
            wait_response: Whether to wait for server response.
            kwargs: RPC arguments.
        """
        selected = self.worker_names
        if worker_names is not None:
            assert len(set(worker_names).difference(selected)) == 0
            selected = worker_names
        if worker_regex is not None:
            selected = [x for x in selected if re.fullmatch(worker_regex, x)]
        if worker_kwargs is not None:
            assert worker_names is not None
            assert worker_regex is None
            assert len(kwargs) == 0
            assert len(worker_names) == len(
                worker_kwargs
            ), f"{len(worker_names)} != {len(worker_kwargs)}"
        else:
            worker_kwargs = [kwargs for _ in selected]

        # connect _MAX_SOCKET_CONCURRENCY sockets at most
        rs = []
        deadline = time.monotonic() + (timeout or 0)
        for j in range(0, len(selected), _MAX_SOCKET_CONCURRENCY):
            sub_rs: List[WorkerControlPanel.Response] = []
            sub_selected = selected[j : j + _MAX_SOCKET_CONCURRENCY]
            sub_worker_kwargs = worker_kwargs[j : j + _MAX_SOCKET_CONCURRENCY]
            for name, kwargs in zip(sub_selected, sub_worker_kwargs):
                address = self.__worker_addresses[name]
                result_fut = self.__requester.async_request(
                    name, address, command, wait_response, **kwargs
                )
                sub_rs.append(
                    WorkerControlPanel.Response(
                        worker_name=name, result=result_fut
                    )
                )

            if not wait_response:
                continue

            bar = range(len(sub_rs))
            if progress:
                try:
                    import tqdm

                    bar = tqdm.tqdm(bar, leave=False)
                except ModuleNotFoundError:
                    pass
            for r, _ in zip(sub_rs, bar):
                if timeout is not None:
                    timeout = max(0, deadline - time.monotonic())
                try:
                    r.result = r.result.result(timeout=timeout)
                except TimeoutError:
                    r.timed_out = True
            rs.extend(sub_rs)
        return rs

    def get_worker_status(self, worker_name) -> WorkerServerStatus:
        """Get status of a connected worker.
        Raises:
            ValueError if worker is not connected.
        """
        try:
            status_str = name_resolve.wait(
                names.worker_status(
                    experiment_name=self.__experiment_name,
                    trial_name=self.__trial_name,
                    worker_name=worker_name,
                ),
                timeout=60,
            )
            status = WorkerServerStatus(status_str)
        except name_resolve.NameEntryNotFoundError:
            status = WorkerServerStatus.LOST
        return status

    def pulse(self):
        return {
            name: self.get_worker_status(name) for name in self.worker_names
        }


@dataclasses.dataclass
class PollResult:
    # Number of total samples and batches processed by the worker. Specifically:
    # - For an actor worker, sample_count = batch_count = number of env.step()-s being executed.
    # - For a policy worker, number of inference requests being handled, versus how many batches were made.
    # - For a trainer worker, number of samples & batches fed into the trainer (typically GPU).
    sample_count: int
    batch_count: int


class Worker:
    """The worker base class that provides general methods and entry point.

    For simplicity, we use a single-threaded pattern in implementing the worker RPC server. Logic
    of every worker are executed via periodical calls to the poll() method, instead of inside
    another thread or process (e.g. the gRPC implementation). A subclass only needs to implement
    poll() without duplicating the main loop.

    The typical code on the worker side is:
        worker = make_worker()  # Returns instance of Worker.
        worker.run()
    and the later is standardized here as:
        while exit command is not received:
            if worker is started:
                worker.poll()
    """

    def __init__(self, server: Optional[WorkerServer] = None):
        """Initializes a worker server.

        Args:
            server: The RPC server API for the worker to register handlers and poll requests.
        """
        self.__running = False
        self.__exiting = False
        self.config = None
        self.__is_configured = False

        self._server = server
        if server is not None:
            server.register_handler("configure", self.configure)
            server.register_handler("reconfigure", self.reconfigure)
            server.register_handler("start", self.start)
            server.register_handler("pause", self.pause)
            server.register_handler("exit", self.exit)
            server.register_handler("interrupt", self.interrupt)
            server.register_handler("ping", lambda: "pong")

        self.logger = logging.getLogger("worker")
        self.__worker_type = None
        self.__worker_index = None
        self.__last_successful_poll_time = None
        self.__worker_info = None

        self._start_time_ns = None

        self.__set_status(WorkerServerStatus.READY)

    def __set_status(self, status: WorkerServerStatus):
        if self._server is not None:
            self.logger.debug(f"Setting worker server status to {status}")
            self._server.set_status(status)

    @property
    def is_configured(self):
        return self.__is_configured

    def _reconfigure(self, **kwargs) -> system_api.WorkerInformation:
        """Implemented by sub-classes."""
        raise NotImplementedError()

    def _configure(self, config) -> system_api.WorkerInformation:
        """Implemented by sub-classes."""
        raise NotImplementedError()

    def _poll(self) -> PollResult:
        """Implemented by sub-classes."""
        raise NotImplementedError()

    def _stats(self) -> Dict[str, Any]:
        """Implemented by sub-classes. For wandb logging only"""
        return {}

    def configure(self, config):
        assert not self.__running
        self.logger.debug("Configuring with: %s", config)

        r = self._configure(config)
        self.__worker_info = r
        self.__worker_type = r.worker_type
        self.__worker_index = r.worker_index
        self.logger = logging.getLogger(r.worker_type + "-worker", "colored")
        if r.host_key is not None:
            self.__host_key(
                names.worker_key(
                    experiment_name=r.experiment_name,
                    trial_name=r.trial_name,
                    key=r.host_key,
                )
            )
        if r.watch_keys is not None:
            keys = (
                [r.watch_keys]
                if isinstance(r.watch_keys, str)
                else r.watch_keys
            )
            self.__watch_keys(
                [
                    names.worker_key(
                        experiment_name=r.experiment_name,
                        trial_name=r.trial_name,
                        key=k,
                    )
                    for k in keys
                ]
            )

        self.__is_configured = True
        self.logger.debug("Configured successfully")

    def reconfigure(self, **kwargs):
        assert not self.__running
        self.__is_configured = False
        self.logger.info(f"Reconfiguring with: {kwargs}")
        self._reconfigure(**kwargs)
        self.__is_configured = True
        self.logger.info("Reconfigured successfully")

    def start(self):
        self.logger.debug("Starting worker")
        self.__running = True
        self.__set_status(WorkerServerStatus.RUNNING)

    def pause(self):
        self.logger.info("Pausing worker")
        self.__running = False
        self.__set_status(WorkerServerStatus.PAUSED)

    def _exit_hook(self, exit_status: WorkerServerStatus):
        logger.warning(f"Exit with {exit_status}, hook not implemented, pass.")

    def exit(self):
        self.logger.info("Exiting worker")
        self._exit_hook(WorkerServerStatus.COMPLETED)
        import torch.distributed as dist

        if dist.is_initialized():
            dist.destroy_process_group()
        self.__set_status(WorkerServerStatus.COMPLETED)
        self.__exiting = True

    def interrupt(self):
        self.logger.info("Worker interrupted by remote control.")
        self._exit_hook(WorkerServerStatus.INTERRUPTED)
        self.__set_status(WorkerServerStatus.INTERRUPTED)
        raise WorkerException(
            worker_name="worker",
            worker_status=WorkerServerStatus.INTERRUPTED,
            scenario="running",
        )

    def run(self):
        self._start_time_ns = time.monotonic_ns()
        self.__last_update_ns = None
        self.logger.debug("Running worker now")
        try:
            while not self.__exiting:
                self._server.handle_requests()
                if not self.__running:
                    time.sleep(0.05)
                    continue
                if not self.__is_configured:
                    raise RuntimeError("Worker is not configured")
                start_time = time.monotonic_ns()
                r = self._poll()
                poll_time = (time.monotonic_ns() - start_time) / 1e9
                wait_seconds = 0.0
                if self.__last_successful_poll_time is not None:
                    # Account the waiting time since the last successful step.
                    wait_seconds = (
                        start_time - self.__last_successful_poll_time
                    ) / 1e9
                self.__last_successful_poll_time = time.monotonic_ns()

                if r.sample_count == r.batch_count == 0:
                    # time.sleep(0.002)
                    pass
                else:
                    now = time.monotonic_ns()
                    if (
                        self.__last_update_ns is not None
                    ):  # Update new stats with 10 seconds frequency.
                        if (now - self.__last_update_ns) / 1e9 >= 10:
                            duration = (
                                time.monotonic_ns() - self._start_time_ns
                            ) / 1e9
                            self.__last_update_ns = now
                    else:
                        self.__last_update_ns = now
        except KeyboardInterrupt:
            self.exit()
        except Exception as e:
            logger.error(f"Worker encountered error {e}", exc_info=True)
            if isinstance(e, WorkerException):
                raise e
            self.__set_status(WorkerServerStatus.ERROR)
            self._exit_hook(WorkerServerStatus.ERROR)
            raise e

    def __host_key(self, key: str):
        self.logger.info(f"Hosting key: {key}")
        name_resolve.add(
            key, "up", keepalive_ttl=15, replace=True, delete_on_exit=True
        )

    def __watch_keys(self, keys: List[str]):
        self.logger.info(f"Watching keys: {keys}")
        name_resolve.watch_names(keys, call_back=self.exit)


class MappingThread:
    """Wrapped of a mapping thread.
    A mapping thread gets from up_stream_queue, process data, and puts to down_stream_queue.
    """

    def __init__(
        self,
        map_fn,
        interrupt_flag,
        upstream_queue,
        downstream_queue: queue.Queue = None,
        cuda_device=None,
    ):
        """Init method of MappingThread for Policy Workers.

        Args:
            map_fn: mapping function.
            interrupt_flag: main thread sets this value to True to interrupt the thread.
            upstream_queue: the queue to get data from.
            downstream_queue: the queue to put data after processing. If None, data will be discarded after processing.
        """
        self.__map_fn = map_fn
        self.__interrupt = interrupt_flag
        self.__upstream_queue = upstream_queue
        self.__downstream_queue = downstream_queue
        self.__thread = threading.Thread(target=self._run, daemon=True)
        self.__cuda_device = cuda_device

    def is_alive(self) -> bool:
        """Check whether the thread is alive.

        Returns:
            alive: True if the wrapped thread is alive, False otherwise.
        """
        return self.__interrupt or self.__thread.is_alive()

    def start(self):
        """Start the wrapped thread."""
        self.__thread.start()

    def join(self):
        """Join the wrapped thread."""
        self.__thread.join()

    def _run(self):
        if self.__cuda_device is not None:
            set_cuda_device(self.__cuda_device)
        while not self.__interrupt:
            self._run_step()

    def _run_step(self):
        try:
            data = self.__upstream_queue.get(timeout=1)
            data = self.__map_fn(data)
            if self.__downstream_queue is not None:
                self.__downstream_queue.put(data)
        except queue.Empty:
            pass

    def stop(self):
        """Stop the wrapped thread."""
        self.__interrupt = True
        if self.__thread.is_alive():
            self.__thread.join()



from typing import List, Optional
import dataclasses

from realhf.api.core.dfg import MFCDef
from realhf.api.quickstart.device_mesh import DeviceMesh
from realhf.api.quickstart.model import ParallelismConfig


@dataclasses.dataclass
class RPCExecution:
    rpc: MFCDef
    device_mesh: DeviceMesh
    parallel_strategy: ParallelismConfig
    time_cost: Optional[int] = None
    mem: Optional[int] = None
    static_mem: Optional[int] = None

    def __repr__(self):
        return f"RPCExecution({self.rpc}, {self.device_mesh}, {self.parallel_strategy})"

    def __hash__(self):
        return hash(
            (
                self.rpc.name,
                self.device_mesh.cluster_mesh,
                self.device_mesh.device_mesh_name,
                str(self.parallel_strategy),
            )
        )


@dataclasses.dataclass
class RPCInstance:
    rpc: MFCDef
    iteration_id: int
    parents: List[MFCDef]
    children: List[MFCDef]

    @property
    def name(self):
        return f"{self.rpc.name}:{self.iteration_id}"

    def __repr__(self):
        if len(self.parents) == 0 and len(self.children) == 0:
            return f"RPCInstance({self.rpc.name}, {self.iteration_id})"
        else:
            return (
                f"RPCInstance({self.rpc.name}, {self.iteration_id}, "
                f"{self.parents}, {self.children})"
            )

    def __hash__(self):
        return hash((self.rpc.name, self.iteration_id))



from dataclasses import dataclass, field
from typing import *
import asyncio
import copy
import time

import numpy as np

import realhf.api.core.dfg as dfg
import realhf.base.logging as logging

logger = logging.getLogger("buffer")


def _extract_intervals(arr):
    if len(arr) == 0:
        return []

    # Initialize the list to hold the intervals
    intervals = []

    # Start of the first interval
    start = arr[0]

    for i in range(1, len(arr)):
        # Check if the current element is not contiguous with the previous one
        if arr[i] != arr[i - 1] + 1:
            # End of the current interval
            end = arr[i - 1]
            # Add the interval as a tuple
            intervals.append((start, end + 1))
            # Start a new interval
            start = arr[i]

    # Add the last interval
    intervals.append((start, arr[-1] + 1))

    return intervals


class BufferFull(Exception):
    pass


@dataclass
class _ReplayEntry:
    reuses_left: int
    receive_time: float
    # We don't save data explicitly, but store metadata.
    # We can know shape and dtype from seqlen and key.
    keys: List[str]
    seqlen: int
    hash_val: int


class _TensorDictSequenceBuffer:
    """An thread-unsafe buffer implementation based on list.

    Used as an internal buffer object in asyncio-based SequenceBuffer.
    Can be replaced with a more efficient C++ implementation based on std vector.

    Methods starting with _ should be called in a locked context.
    """

    def __init__(self, keys: List[str], max_size: int, reuses: int):
        # Fixed-size storage, storing pointers, but sequeces in dict have variable lengths.
        self.__storage: List[_ReplayEntry] = [None for _ in range(max_size)]

        # Some states of the storage. Read/Write applied to them should be locked.
        self.__seqlens = np.zeros(max_size, dtype=np.int32)
        self.__hash_vals = np.zeros(max_size, dtype=np.int64)
        self.__has_keys = np.zeros((max_size, len(keys)), dtype=bool)

        self.__keys = keys
        self.__reuses = reuses

    def _update_seqlen(self, indices: int):
        self.__seqlens[indices] = [
            self.__storage[idx].seqlen for idx in indices
        ]

    def _get_seqlen(self, indices: int) -> np.ndarray:
        return self.__seqlens[indices]

    def _update_hash_val(self, indices: int):
        self.__hash_vals[indices] = [
            self.__storage[idx].hash_val for idx in indices
        ]

    def _get_hash_val(self, indices: int) -> np.ndarray:
        return self.__hash_vals[indices]

    def _update_has_keys(self, indices: List[int]):
        for idx in indices:
            self.__has_keys[idx] = [
                k in self.__storage[idx].keys for k in self.__keys
            ]

    def _get_has_keys(self, indices):
        return self.__has_keys[indices, :]

    def put_batch(
        self, indices: List[int], xs: List[Tuple[List[str], int, int]]
    ):
        assert len(indices) == len(xs)
        # Can be parallelized.
        for idx, x in zip(indices, xs):
            keys, seqlen, hash_val = x
            self.__storage[idx] = _ReplayEntry(
                reuses_left=self.__reuses,
                receive_time=time.time(),
                keys=copy.deepcopy(keys),
                seqlen=seqlen,
                hash_val=hash_val,
            )

    def amend_batch(
        self, indices: List[int], new_datas: List[Tuple[List[str], int]]
    ):
        assert len(indices) == len(new_datas)
        # Can be parallelized.
        for idx, new_data in zip(indices, new_datas):
            new_keys, new_seqlen = new_data
            assert (
                len(set(new_keys).intersection(self.__storage[idx].keys)) == 0
            ), (
                new_keys,
                self.__storage[idx].keys,
            )
            self.__storage[idx].keys += new_keys
            self.__storage[idx].seqlen = new_seqlen

    def get_batch(self, indices: List[int]) -> List[_ReplayEntry]:
        # Can be parallelized.
        res = []
        for idx in indices:
            r = self.__storage[idx]
            r.reuses_left -= 1
            res.append(r)
        return res

    def pop_batch(self, indices: List[int]):
        res = []
        for idx in indices:
            r = self.__storage[idx]
            self.__storage[idx] = None
            self.__seqlens[idx] = 0
            self.__hash_vals[idx] = 0
            self.__has_keys[idx] = False
            res.append(r)
        return res


@dataclass
class SequenceSample:
    indices: List[int]
    seqlens: List[int]
    hash_vals: List[int]


class AsyncIOSequenceBuffer:

    def __init__(
        self,
        rpcs: List[dfg.MFCDef],
        max_size: int,
        fetch_ctl: asyncio.Queue,
        fetch_master_ctl: asyncio.Queue,
    ):
        self._lock = asyncio.Condition(asyncio.Lock())

        # Both are queues of size 1.
        self._fetch_ctl = fetch_ctl
        self._fetch_master_ctl = fetch_master_ctl
        self._load_data_requested = False

        # Buffer indicators, should be locked by self._lock.
        # Put, amend, ready, idle, and empty are mutually exclusive.
        self._is_being_put = np.zeros(max_size, dtype=bool)
        self._is_being_amended = np.zeros(max_size, dtype=bool)
        self._is_being_read = np.zeros(max_size, dtype=bool)
        self._is_idle = np.zeros(max_size, dtype=bool)
        self._is_empty = np.ones(max_size, dtype=bool)
        self._buf_size = 0
        self._n_tokens = 0
        # We allow concurrent amenders and readers.
        self._n_amenders = np.zeros(max_size, dtype=int)
        self._n_readers = np.zeros(max_size, dtype=int)

        self._ready_for_rpcs = np.zeros((max_size, len(rpcs)), dtype=bool)
        self._completed_rpc = np.zeros((max_size, len(rpcs)), dtype=bool)

        self._rpc_data_keys = rpc_data_keys = list(
            set().union(*[rpc.input_data for rpc in rpcs])
        )
        # We can efficiently compute whether an RPC is ready using this mask
        self._rpc_key_mask = np.stack(
            [
                np.array(
                    [k in rpc.input_data for k in rpc_data_keys], dtype=bool
                )
                for rpc in rpcs
            ],
            axis=1,
        )
        self._rpc_names = [rpc.name for rpc in rpcs]

        # The internal buffer implementation.
        self.__max_size = max_size
        self.__buffer = _TensorDictSequenceBuffer(
            keys=rpc_data_keys, max_size=max_size, reuses=len(rpcs)
        )

    @property
    def lock(self):
        return self._lock

    @property
    def n_rpcs(self):
        return len(self._rpc_names)

    def _assert_valid_indicator(self):
        assert (
            self._is_being_put
            + self._is_being_amended
            + self._is_being_read
            + self._is_idle
        ).sum() == self._buf_size
        assert (self._is_empty.sum() + self._buf_size) == self.__max_size
        assert ((self._n_amenders > 0) == self._is_being_amended).all()
        assert (self._n_amenders >= 0).all()
        assert ((self._n_readers > 0) == self._is_being_read).all()
        assert (self._n_readers >= 0).all()
        assert (self._is_empty[:, None] * self._ready_for_rpcs).sum() == 0
        assert (self._is_empty[:, None] * self._completed_rpc).sum() == 0

    async def put_batch(self, samples: List[Tuple[List[str], int, int]]):
        async with self._lock:
            self._assert_valid_indicator()
            n = len(samples)
            indices = np.where(self._is_empty)[0]
            indices = np.arange(*_extract_intervals(indices)[-1])[:n]
            if len(indices) < n:
                raise BufferFull("Please set a larger buffer size")
            self._is_empty[indices] = False
            self._is_being_put[indices] = True

        self.__buffer.put_batch(indices, samples)

        async with self._lock:
            self.__buffer._update_has_keys(indices)
            self.__buffer._update_seqlen(indices)
            self.__buffer._update_hash_val(indices)

            has_keys = self.__buffer._get_has_keys(indices)  # [bs, #keys]
            rpc_key_mask = self._rpc_key_mask  # [#keys, #rpcs]
            self._ready_for_rpcs[indices] = (
                has_keys[:, :, None] >= rpc_key_mask[None, :, :]
            ).all(axis=1)

            self._is_being_put[indices] = False
            self._is_idle[indices] = True

            self._buf_size += len(samples)
            self._n_tokens += self.__buffer._get_seqlen(indices).sum()
            if self._buf_size >= 0.95 * self.__max_size:
                logger.warning(
                    f"Buffer is 95% full. The current buffer size is {self._buf_size} "
                    f"while the maximum size is {self.__max_size}. "
                    f"If your dataset has more than 1M sequences, consider enlarge "
                    f"the default batch size in the master worker."
                )
            self._load_data_requested = False
        return indices

    async def amend_batch(
        self, indices: List[int], new_datas: List[Tuple[List[str], int, int]]
    ):
        async with self._lock:
            await self._lock.wait_for(
                lambda: (
                    self._is_idle[indices] | self._is_being_amended[indices]
                ).all(),
            )
            self._assert_valid_indicator()
            self._is_idle[indices] = False
            self._is_being_amended[indices] = True
            self._n_amenders[indices] += 1
        self.__buffer.amend_batch(indices, new_datas)
        async with self._lock:
            self._n_tokens -= self.__buffer._get_seqlen(indices).sum()
            self.__buffer._update_has_keys(indices)
            self.__buffer._update_seqlen(indices)
            self.__buffer._update_hash_val(indices)
            self._n_tokens += self.__buffer._get_seqlen(indices).sum()

            has_keys = self.__buffer._get_has_keys(indices)  # [bs, #keys]
            rpc_key_mask = self._rpc_key_mask  # [#keys, #rpcs]
            self._ready_for_rpcs[indices] = (
                has_keys[:, :, None] >= rpc_key_mask[None, :, :]
            ).all(axis=1)

            self._n_amenders[indices] -= 1
            self._is_being_amended[indices] = self._n_amenders[indices] > 0
            self._is_idle[indices] = np.logical_not(
                self._is_being_amended[indices]
            )
            if self._is_idle[indices].any():
                self._lock.notify(len(self._rpc_names))

    def _request_load_data(self):
        if not self._load_data_requested:
            try:
                self._fetch_ctl.put_nowait(1)
            except asyncio.QueueFull:
                pass
            try:
                self._fetch_master_ctl.put_nowait(1)
            except asyncio.QueueFull:
                pass
            self._load_data_requested = True

    async def get_batch_for_rpc(self, rpc: dfg.MFCDef) -> SequenceSample:
        rpc_idx = self._rpc_names.index(rpc.name)

        def _can_do_rpc() -> bool:
            ready_indices = np.nonzero(
                (self._is_idle | self._is_being_read)
                & self._ready_for_rpcs[:, rpc_idx]
                & ~self._completed_rpc[:, rpc_idx]
            )[0]
            if len(ready_indices) < rpc.min_n_seqs:
                return False
            return True

        async with self._lock:
            # await self._lock.wait_for(_can_do_rpc)
            if rpc.is_src:
                ready_indices = np.nonzero(
                    (self._is_idle | self._is_being_read)
                    & self._ready_for_rpcs[:, rpc_idx]
                    & ~self._completed_rpc[:, rpc_idx]
                )[0]
                # *2 because we want to fetch new data as long as the *next* RPC does not have enough data.
                if len(ready_indices) < rpc.min_n_seqs * 2:
                    self._request_load_data()

            while not _can_do_rpc():
                await self._lock.wait()

            self._assert_valid_indicator()

            ready_indices = np.nonzero(
                (self._is_idle | self._is_being_read)
                & self._ready_for_rpcs[:, rpc_idx]
                & ~self._completed_rpc[:, rpc_idx]
            )[0]
            seqlens = self.__buffer._get_seqlen(ready_indices)
            hash_vals = self.__buffer._get_hash_val(ready_indices)

            indices = ready_indices[: rpc.max_n_seqs]
            seqlens = seqlens[: rpc.max_n_seqs]
            hash_vals = hash_vals[: rpc.max_n_seqs]
            assert rpc.min_n_seqs <= len(indices) <= rpc.max_n_seqs, (
                rpc.min_n_seqs,
                len(indices),
                rpc.max_n_seqs,
            )

            self._is_idle[indices] = False
            self._is_being_read[indices] = True
            self._n_readers[indices] += 1

        entries = self.__buffer.get_batch(indices)
        assert all([entry.reuses_left >= 0 for entry in entries])
        pop_indices = [
            idx
            for idx, entry in zip(indices, entries)
            if entry.reuses_left == 0
        ]
        # The following call is safe because no more RPC will write to popped data.
        pop_tokens = self.__buffer._get_seqlen(pop_indices).sum()
        if len(pop_indices) > 0:
            self.__buffer.pop_batch(pop_indices)

        async with self._lock:
            self._n_readers[indices] -= 1
            self._is_being_read[indices] = self._n_readers[indices] > 0
            self._is_idle[indices] = self._n_readers[indices] == 0
            self._completed_rpc[indices, rpc_idx] = True

            assert (self._n_readers[pop_indices] == 0).all()
            assert (self._n_amenders[pop_indices] == 0).all()
            self._is_empty[pop_indices] = True
            self._is_idle[pop_indices] = False
            self._completed_rpc[pop_indices] = False
            self._ready_for_rpcs[pop_indices] = False
            self._buf_size -= len(pop_indices)
            self._n_tokens -= pop_tokens

            if self._is_idle[indices].any():
                self._lock.notify(len(self._rpc_names))
        return SequenceSample(
            indices=indices, seqlens=seqlens, hash_vals=hash_vals
        )



from typing import Type
import importlib
import os
import traceback

import realhf.api.core.system_api
import realhf.base.logging as logging

logger = logging.getLogger("system")

# NOTE: Workers are configured in the following order.
# Take special care when adding a new worker type.
WORKER_TYPES = ["model_worker", "master_worker"]


def load_worker(worker_type: str) -> Type:
    assert worker_type in WORKER_TYPES, f"Invalid worker type {worker_type}"
    module = importlib.import_module(worker_type_to_module(worker_type))
    class_name = worker_type_to_class_name(worker_type)
    return getattr(module, class_name)


def worker_type_to_module(worker_type: str):
    return "realhf.system." + worker_type


def worker_type_to_class_name(worker_type: str):
    return "".join([w.capitalize() for w in worker_type.split("_")])


def run_worker(
    worker_type, experiment_name, trial_name, worker_name, worker_server_type
):
    """Run one worker
    Args:
        worker_type: string, one of the worker types listed above,
        experiment_name: string, the experiment this worker belongs to,
        trial_name: string, the specific trial this worker belongs to,
        worker_name: name given to the worker, typically "<worker_type>/<worker_index>"
        worker_server_type: string, either 'zmq' or 'ray'.
    """
    worker_class = load_worker(worker_type)
    make_server_fn = getattr(
        importlib.import_module("realhf.system.worker_control"), "make_server"
    )
    server = make_server_fn(
        type_=worker_server_type,
        experiment_name=experiment_name,
        trial_name=trial_name,
        worker_name=worker_name,
    )
    worker = worker_class(server=server)
    try:
        worker.run()
    except Exception as e:
        logger.error("Worker %s failed with exception: %s", worker_name, e)
        logger.error(traceback.format_exc())
        raise e


def make_controller(type_, experiment_name, trial_name):
    module = importlib.import_module("realhf.system.controller")
    if type_ == "zmq":
        control_module = importlib.import_module("realhf.system.worker_control")
        panel = getattr(control_module, "make_control")(
            "zmq", experiment_name, trial_name
        )
        return getattr(module, "Controller")(experiment_name, trial_name, panel)
    elif type_ == "ray":
        return getattr(module, "RayController")(
            experiment_name, trial_name, local_mode=False
        )
    elif type_ == "local_ray":
        return getattr(module, "RayController")(
            experiment_name, trial_name, local_mode=True
        )
    else:
        raise NotImplementedError(f"Unknown controller type {type_}.")



from typing import Dict, List, Optional, Tuple, Union
import abc
import collections

import numpy as np
import torch

import realhf.base.datapack as datapack
import realhf.base.logging as logging
import realhf.base.namedarray as namedarray

logger = logging.getLogger("dataparallel")

InterfaceReturnDataType = Union[namedarray.NamedArray, Dict]


class ParallelDataBroker:

    @staticmethod
    def gather_from(
        src: List[InterfaceReturnDataType],
    ) -> InterfaceReturnDataType:
        """Gather outputs of a data-parallel model RPC."""
        if isinstance(src[0], Dict):
            cnt, stats = {}, {}
            for reply in src:
                for k, v in reply.items():
                    cnt[k] = cnt.get(k, 0) + 1
                    stats[k] = stats.get(k, 0) + v
            return {
                k: v / cnt
                for k, v, cnt in zip(stats.keys(), stats.values(), cnt.values())
            }
        elif isinstance(src[0], namedarray.NamedArray):
            if "input_lens" in src[0]:
                input_lens = torch.cat([x["input_lens"] for x in src], dim=0)
            elif "cu_seqlens" in src[0]:
                input_lens = torch.cat(
                    [x["cu_seqlens"][1:] - x["cu_seqlens"][:-1] for x in src],
                    dim=0,
                )
            elif "prompt_lens" in src[0]:
                input_lens = torch.cat([x["prompt_lens"] for x in src], dim=0)
            elif "prompt_cu_seqlens" in src[0]:
                input_lens = torch.cat(
                    [
                        x["prompt_cu_seqlens"][1:] - x["prompt_cu_seqlens"][:-1]
                        for x in src
                    ],
                    dim=0,
                )
            res = namedarray.recursive_aggregate(
                src, lambda x: torch.cat(x, dim=0)
            )
            if "cu_seqlens" in src[0] and len(src[0]["cu_seqlens"].shape) == 1:
                res["cu_seqlens"] = torch.cat(
                    [input_lens.new_zeros(1), torch.cumsum(input_lens, dim=0)]
                )
            elif (
                "prompt_cu_seqlens" in src[0]
                and len(src[0]["prompt_cu_seqlens"].shape) == 1
            ):
                res["prompt_cu_seqlens"] = torch.cat(
                    [input_lens.new_zeros(1), torch.cumsum(input_lens, dim=0)]
                )
            return res
        else:
            raise NotImplementedError(
                f"Don't know how to gather data of type {[type(x) for x in src]}."
            )

    @abc.abstractstaticmethod
    def scatter_to(
        src: namedarray.NamedArray, n_dp: int, **kwargs
    ) -> List[namedarray.NamedArray]:
        """Scatter the input of a data-parallel model RPC."""
        pass


class PaddedBatchParallelDataBroker(ParallelDataBroker):

    @staticmethod
    def gather_from(
        src: List[InterfaceReturnDataType],
    ) -> InterfaceReturnDataType:
        return ParallelDataBroker.gather_from(src)

    @staticmethod
    def scatter_to(
        src: namedarray.NamedArray, n_dp: int, **kwargs
    ) -> List[namedarray.NamedArray]:
        datas = namedarray.split(src, n_dp)
        for x in datas:
            x.register_metadata(**src.metadata)
        return datas


class PackedParallelDataBroker(ParallelDataBroker):

    @staticmethod
    def gather_from(
        src: List[InterfaceReturnDataType],
    ) -> InterfaceReturnDataType:
        return ParallelDataBroker.gather_from(src)

    @staticmethod
    def scatter_to(
        src: namedarray.NamedArray,
        n_dp: int,
        return_sizes=False,
        partitions: Optional[List[Tuple[int, int]]] = None,
        min_size: int = 1,
        return_partitions=False,
    ) -> List[namedarray.NamedArray]:
        if "input_lens" not in src:
            if "cu_seqlens" in src:
                raw_input_lens = src["cu_seqlens"][1:] - src["cu_seqlens"][:-1]
            elif "prompt_lens" in src or "prompt_cu_seqlens" in src:
                if "prompt_lens" in src:
                    raw_input_lens = src["prompt_lens"]
                else:
                    raw_input_lens = (
                        src["prompt_cu_seqlens"][1:]
                        - src["prompt_cu_seqlens"][:-1]
                    )
            else:
                raise RuntimeError(
                    "input_lens/cu_seqlens/prompt_lens/prompt_cu_seqlens "
                    "must be in the return data when using packed data broker. "
                    f"Current keys: {list(src.keys())}."
                )
        else:
            raw_input_lens = src["input_lens"]
        assert src.metadata.get("seqlens", None) is not None
        if "seqlens" in src.metadata:
            seqlens_cpu = src.metadata["seqlens"]
            assert isinstance(seqlens_cpu, list)
            seqlens_cpu = np.array(seqlens_cpu, dtype=np.int64)
        else:
            logger.warning(
                "seqlens not found in metadata. Using input_lens to calculate partitions. "
                "This will create an unnecessary host-device synchronization point."
            )
            seqlens_cpu = raw_input_lens.cpu().numpy().astype(np.int64)

        if partitions is None:
            partitions = datapack.min_abs_diff_partition(
                seqlens_cpu, n_dp, min_size
            )

        input_lens: List[torch.IntTensor] = [
            raw_input_lens[start:end].int() for start, end in partitions
        ]
        cu_seqlens = [
            torch.cat([x.new_zeros(1), torch.cumsum(x, dim=0)]).int()
            for x in input_lens
        ]
        batch_sizes = [cu_seqlen.shape[0] - 1 for cu_seqlen in cu_seqlens]

        # These are used by log probabilities, which are one-step shorter than packed inputed ids.
        # We use numpy/list for indexing to avoid host-device synchronization
        batch_seqlens = [
            sum(seqlens_cpu[start:end]) for start, end in partitions
        ]
        offsets = [0] + np.cumsum(batch_seqlens).tolist()[:-1]
        short1batch_seqlens = [
            sum(seqlens_cpu[start:end]) - (end - start)
            for start, end in partitions
        ]
        short1offsets = [0] + np.cumsum(short1batch_seqlens).tolist()[:-1]

        splitted_data = [collections.defaultdict() for _ in range(n_dp)]
        for k, v in src.items():
            for i, sp in enumerate(splitted_data):
                # NOTE: This is a terrible implementation, because we must know how to split each tensor according to its semantics.
                # Different tensor has different shape and semantics,
                # e.g., packed_seq has shape [tot_seqlen] and should be splited according to cumulative lengths,
                # packed_logprobs has shape [tot_seqlen - bs] (each sequence is one-step shorter) and should be splitted
                # according to short-1 cumulative lengths, seq_no_eos_mask has shape [bs] and performs similar as input_lens, ...
                # so we must enumerate each possible key and deal with them separately, etc.
                if v is None:
                    sp[k] = None
                elif k in ["prompt_cu_seqlens", "cu_seqlens"]:
                    continue
                elif k in [
                    "prompt_lens",
                    "input_lens",
                    "seq_no_eos_mask",
                    "rewards",
                    "reward_score",
                    "group_factor",
                    "pos_input_lens",
                    "group_input_lens",
                    "seqlogp",
                ]:
                    start, end = partitions[i]
                    sp[k] = v[start:end]
                elif k in [
                    "packed_seq",
                    "packed_logits_mask",
                    "prompt_mask",
                    "packed_input_ids",
                    "values",
                    "logits_mask",
                    "packed_prompts",
                ]:
                    sp[k] = v[offsets[i] : offsets[i] + batch_seqlens[i]]
                elif k in [
                    "packed_logprobs",
                    "packed_ref_logprobs",
                    "old_logp",
                    "ref_logp",
                    "advantages",
                    "ppo_loss_mask",
                    "kl_rewards",
                    "returns",
                ]:
                    sp[k] = v[
                        short1offsets[i] : short1offsets[i]
                        + short1batch_seqlens[i]
                    ]
                elif not torch.is_tensor(src[k]):
                    # for constant, preserve value for each splitted instance
                    sp[k] = src[k]
                else:
                    raise RuntimeError(
                        f"Unknown key {k} in packed data. We don't know how to split it. "
                        f"Check base/dataparallel.py for implemented keys."
                    )
        if "cu_seqlens" in src.keys():
            for x, slens in zip(splitted_data, input_lens):
                x["cu_seqlens"] = torch.nn.functional.pad(
                    slens.cumsum(dim=0), (1, 0)
                ).int()
        if "prompt_cu_seqlens" in src.keys():
            raw_prompt_lens = (
                src["prompt_cu_seqlens"][1:] - src["prompt_cu_seqlens"][:-1]
            )
            all_prompt_lens: List[torch.IntTensor] = [
                raw_prompt_lens[start:end].int() for start, end in partitions
            ]
            for x, pslens in zip(splitted_data, all_prompt_lens):
                x["prompt_cu_seqlens"] = torch.nn.functional.pad(
                    pslens.cumsum(dim=0), (1, 0)
                ).int()
        splitted_data = [namedarray.from_dict(dict(x)) for x in splitted_data]
        for x in splitted_data:
            x.register_metadata(**src.metadata)

        res = [splitted_data]
        if return_sizes:
            res.append(batch_sizes)
        if return_partitions:
            res.append(partitions)
        if not return_sizes and not return_partitions:
            return res[0]
        else:
            return res


def get_broker(type_: str) -> ParallelDataBroker:
    if type_ == "padded_batch":
        return PaddedBatchParallelDataBroker
    elif type_ == "packed":
        return PackedParallelDataBroker
    else:
        raise RuntimeError(f"Unknown data broker type {type_}.")



import argparse
import functools
import json
import multiprocessing
import os
import pickle
import re
import socket
import subprocess

from omegaconf import OmegaConf
import torch

multiprocessing.set_start_method("spawn", force=True)

from realhf.api.quickstart.entrypoint import (
    QUICKSTART_CONFIG_CLASSES,
    QUICKSTART_EXPR_CACHE_PATH,
)
from realhf.base import (
    cluster,
    gpu_utils,
    importing,
    logging,
    name_resolve,
    names,
)

RAY_HEAD_WAIT_TIME = 500
logger = logging.getLogger("Main-Workers")


def _patch_external_impl(exp_name, trial_name):
    import realhf.api.core.system_api as system_api

    if os.path.exists(QUICKSTART_EXPR_CACHE_PATH):
        for exp_cache in os.listdir(QUICKSTART_EXPR_CACHE_PATH):
            target_cache_name = f"{exp_name}_{trial_name}.json"
            if exp_cache != target_cache_name:
                continue
            cache_file = os.path.join(
                QUICKSTART_EXPR_CACHE_PATH, target_cache_name
            )
            with open(cache_file, "r") as f:
                cache = json.load(f)
            usercode_path = cache["usercode_path"]
            exp_cls_args = OmegaConf.create(cache["args"])
            config_name = cache["config_name"]
            # Import user code to register quickstart experiments.
            importing.import_usercode(usercode_path, "_realhf_user_code")
            # Register the internal experiment.
            exp_cls = QUICKSTART_CONFIG_CLASSES[config_name]
            system_api.register_experiment(
                exp_name, functools.partial(exp_cls, **exp_cls_args)
            )


def main_worker(args):
    import realhf.base.constants as constants

    constants.set_experiment_trial_names(args.experiment_name, args.trial_name)
    _patch_external_impl(args.experiment_name, args.trial_name)

    worker_index_start = (
        args.jobstep_id * args.wprocs_per_jobstep + args.wproc_offset
    )
    worker_index_end = min(
        worker_index_start + args.wprocs_per_jobstep,
        args.wprocs_in_job + args.wproc_offset,
    )

    group_name = f"{args.worker_type}-{args.worker_submission_index}"
    logger.debug(
        f"{args.worker_type} group id {args.jobstep_id}, "
        f"worker index {worker_index_start}:{worker_index_end}"
    )

    # Isolate within the same slurm job, among different jobsteps.
    if torch.cuda.is_initialized():
        raise RuntimeError(
            "CUDA already initialized before isolating CUDA devices. This should not happen."
        )
    gpu_utils.isolate_cuda_device(
        group_name,
        args.jobstep_id,
        args.n_jobsteps,
        args.experiment_name,
        args.trial_name,
    )
    if os.environ.get("CUDA_VISIBLE_DEVICES", None):
        logger.debug(
            "CUDA_VISIBLE_DEVICES: %s", os.environ["CUDA_VISIBLE_DEVICES"]
        )

    # NOTE: Importing these will initialize DeepSpeed/CUDA devices.
    # profiler.import_profiler_registers()
    import realhf.impl.dataset
    import realhf.impl.model
    import realhf.system

    logger.debug(f"Run {args.worker_type} worker with args: %s", args)
    assert not args.experiment_name.startswith(
        "/"
    ), f'Invalid experiment_name "{args.experiment_name}" starts with "/"'
    if args.wprocs_per_jobstep == 1:
        realhf.system.run_worker(
            worker_type=args.worker_type,
            experiment_name=args.experiment_name,
            trial_name=args.trial_name,
            worker_name=f"{args.worker_type}/{worker_index_start}",
            worker_server_type="zmq",
        )
    else:
        workers = []
        for wid in range(worker_index_start, worker_index_end):
            worker_args = dict(
                worker_type=args.worker_type,
                experiment_name=args.experiment_name,
                trial_name=args.trial_name,
                worker_name=f"{args.worker_type}/{wid}",
                worker_server_type="zmq",
            )
            p = multiprocessing.Process(
                target=realhf.system.run_worker, kwargs=worker_args
            )
            p.name = f"{args.worker_type}/{wid}"
            p.start()
            workers.append(p)

        logger.info(
            f"Waiting for {args.wprocs_per_jobstep} {args.worker_type} workers of group id {args.jobstep_id}."
        )
        worker_exit_code = 0
        while not worker_exit_code:
            for p in workers:
                if p.exitcode is None:
                    p.join(timeout=5)
                elif p.exitcode == 0:
                    pass
                else:
                    logger.error(f"{p.name} exitcode: {p.exitcode}")
                    worker_exit_code = p.exitcode
                    break
        for p in workers:
            if p.is_alive():
                p.kill()
        exit(worker_exit_code)


def main_controller(args):
    """
    Args:
        args: argparse result including:
            experiment_name:
            trial_name:
            config_index: the index of experiment configuration (experiment may return multiple configurations)
            ignore_worker_error: bool, if False, stop the experiment when any worker(s) fail.
    """
    import realhf.api.core.system_api as system_api
    import realhf.base.constants as constants
    import realhf.system as system

    constants.set_experiment_trial_names(args.experiment_name, args.trial_name)
    _patch_external_impl(args.experiment_name, args.trial_name)

    logger.debug("Running controller with args: %s", args)
    assert not args.experiment_name.startswith("/"), args.experiment_name
    if args.type == "ray":
        # launch ray cluster head
        ray_flags = [
            f"--num-cpus=0",
            f"--num-gpus=0",
            f"--port={args.ray_port}",
            "--head",
        ]
        cmd = f"ray start {' '.join(ray_flags)}"
        output = subprocess.check_output(cmd, shell=True).decode("ascii")
        logger.info("Successfully launched ray cluster head.")

        pattern = r"ray start --address='(\d+\.\d+\.\d+\.\d+:\d+)'"
        match = re.search(pattern, output)
        if match:
            addr = match.group(1)
            logger.info("Found ray address: '%s'", addr)
        else:
            raise RuntimeError(
                f"Address not found in ray start output: {output}."
            )
        ray_addr_name = names.ray_cluster(
            args.experiment_name, args.trial_name, "address"
        )
        name_resolve.add(
            ray_addr_name,
            addr,
            delete_on_exit=True,
            keepalive_ttl=RAY_HEAD_WAIT_TIME,
        )

    controller = system.make_controller(
        type_=args.type,
        experiment_name=args.experiment_name,
        trial_name=args.trial_name,
    )
    experiment = system_api.make_experiment(args.experiment_name)
    controller.start(
        experiment=experiment,
        ignore_worker_error=args.ignore_worker_error,
    )

    if args.type == "ray":
        subprocess.check_output(f"ray stop", shell=True)


def main_ray(args):
    ray_addr_name = names.ray_cluster(
        args.experiment_name, args.trial_name, "address"
    )
    try:
        address = name_resolve.wait(ray_addr_name, timeout=RAY_HEAD_WAIT_TIME)
    except TimeoutError:
        raise TimeoutError("Timeout waiting for ray cluster head address.")
    ray_flags = [f"--address={address}"]

    cmd = f"ray start {' '.join(ray_flags)}"
    _ = subprocess.check_output(cmd, shell=True).decode("ascii")
    logger.info(
        f"Successfully launched nodes for {args.worker_type} in Ray cluster."
    )

    host_ip = socket.gethostbyname(socket.gethostname())
    name_resolve.add(
        names.ray_cluster(
            args.experiment_name,
            args.trial_name,
            f"{args.worker_type}/{args.jobstep_id}",
        ),
        host_ip,
        delete_on_exit=True,
        keepalive_ttl=300,
    )

    while True:
        try:
            ray_exiting_name = names.ray_cluster(
                args.experiment_name, args.trial_name, "exiting"
            )
            try:
                name_resolve.wait(ray_exiting_name, timeout=10)
                break
            except TimeoutError:
                pass
        except KeyboardInterrupt:
            break

    subprocess.check_output(f"ray stop", shell=True)


def main():
    parser = argparse.ArgumentParser(prog="marl")
    subparsers = parser.add_subparsers(dest="cmd", help="sub-command help")
    subparsers.required = True

    subparser = subparsers.add_parser(
        "controller", help="run a controller of experiment"
    )
    subparser.add_argument("--experiment_name", "-e", type=str, required=True)
    subparser.add_argument("--trial_name", "-f", type=str, required=True)
    subparser.add_argument("--ignore_worker_error", action="store_true")
    subparser.add_argument(
        "--raise_worker_error", dest="ignore_worker_error", action="store_false"
    )
    subparser.add_argument("--type", type=str, default="zmq")
    subparser.add_argument("--ray_port", type=int, default=8777)
    subparser.set_defaults(feature=False)
    subparser.set_defaults(func=main_controller)

    subparser = subparsers.add_parser("worker", help="run a standalone worker")
    subparser.add_argument("--experiment_name", "-e", type=str, required=True)
    subparser.add_argument("--trial_name", "-f", type=str, required=True)
    subparser.add_argument("--worker_type", "-w", type=str, required=True)
    subparser.add_argument(
        "--jobstep_id",
        "-i",
        type=int,
        required=True,
        help="jobstep/task ID in a slurm job.",
    )
    subparser.add_argument(
        "--n_jobsteps",
        "-g",
        type=int,
        required=True,
        help="`--ntasks` of `srun`, aka SLURM_NPROCS.",
    )
    subparser.add_argument(
        "--worker_submission_index",
        "-r",
        type=int,
        required=True,
        help="Submission index to slurm for this worker. Used for locating job name and logs.",
    )
    subparser.add_argument(
        "--wprocs_per_jobstep",
        "-p",
        type=int,
        required=True,
        help="Number of worker processes launched by multiprocessing in this script.",
    )
    subparser.add_argument(
        "--wprocs_in_job",
        "-j",
        type=int,
        required=True,
        help="Number of worker processes in this slurm job.",
    )
    subparser.add_argument(
        "--wproc_offset",
        "-o",
        type=int,
        required=True,
        help="Offset of worker processes of this slurm job. "
        "For example, we may allocate 4 type `A` workers with 1 GPU each and 2 with 0.5 GPU each. "
        "This launches 2 jobs, the former with 4 job steps and the latter with 2 job steps. "
        "The offset is 0 for the 1st job and 4 for the 2nd job.",
    )
    subparser.set_defaults(func=main_worker)

    subparser = subparsers.add_parser(
        "ray", help="launch ray cluster write ray address to name_resolve"
    )
    subparser.add_argument("--experiment_name", "-e", type=str, required=True)
    subparser.add_argument("--trial_name", "-f", type=str, required=True)
    subparser.add_argument("--worker_type", "-w", type=str, required=True)
    subparser.add_argument("--jobstep_id", "-i", type=int, required=True)
    subparser.add_argument("--n_jobsteps", "-g", type=int, required=True)
    subparser.set_defaults(func=main_ray)

    args = parser.parse_args()

    args.func(args)


if __name__ == "__main__":
    main()



from typing import *

import torch
import transformers

from realhf.api.core.model_api import (
    ReaLModelConfig,
    register_hf_family,
    register_hf_path,
)
from realhf.base.constants import use_te_impl


def convert_config_starcoder(
    starcoder_config: transformers.GPTBigCodeConfig,
) -> ReaLModelConfig:
    return ReaLModelConfig(
        n_layers=starcoder_config.n_layer,
        n_kv_heads=1,
        attn_pdrop=starcoder_config.attn_pdrop,
        embd_pdrop=starcoder_config.embd_pdrop,
        layer_norm_epsilon=starcoder_config.layer_norm_epsilon,
        hidden_dim=starcoder_config.n_embd,
        head_dim=starcoder_config.n_embd // starcoder_config.n_head,
        intermediate_dim=starcoder_config.n_inner,
        n_positions=starcoder_config.n_positions,
        resid_pdrop=starcoder_config.resid_pdrop,
        vocab_size=starcoder_config.vocab_size,
    )


def to_starcoder_config(
    config: ReaLModelConfig,
) -> transformers.GPTBigCodeConfig:
    return transformers.GPTBigCodeConfig(
        vocab_size=config.vocab_size,
        n_positions=config.n_positions,
        n_embd=config.hidden_dim,
        n_layer=config.n_layers,
        n_head=config.hidden_dim // config.head_dim,
        n_inner=config.intermediate_dim,
        resid_pdrop=config.resid_pdrop,
        embd_pdrop=config.embd_pdrop,
        attn_pdrop=config.attn_pdrop,
        layer_norm_epsilon=config.layer_norm_epsilon,
        multi_query=(config.n_kv_heads == 1),
        activation_function="gelu",
    )


def _starcoder_key_mapping_fn(config: ReaLModelConfig) -> Dict[str, str]:
    assert not use_te_impl(), "starcoder does not support TE backend now"
    key_mappings = {
        "transformer.wte.weight": "0.wte.weight",
        "transformer.wpe.weight": "0.wpe.weight",
        "transformer.ln_f.weight": f"{config.n_layers}.ln_f.weight",
        "transformer.ln_f.bias": f"{config.n_layers}.ln_f.bias",
        "lm_head.weight": f"{config.n_layers+1}.weight",
    }
    for i in range(config.n_layers):
        key_mappings[f"transformer.h.{i}.ln_1.weight"] = (
            f"{i + 1}.attn.c_attn.ln.weight"
        )
        key_mappings[f"transformer.h.{i}.ln_1.bias"] = (
            f"{i + 1}.attn.c_attn.ln.bias"
        )
        key_mappings[f"transformer.h.{i}.attn.c_attn.weight"] = (
            f"{i + 1}.attn.c_attn.linear.weight"
        )
        key_mappings[f"transformer.h.{i}.attn.c_attn.bias"] = (
            f"{i + 1}.attn.c_attn.linear.bias"
        )
        key_mappings[f"transformer.h.{i}.attn.c_proj.weight"] = (
            f"{i + 1}.attn.c_proj.weight"
        )
        key_mappings[f"transformer.h.{i}.attn.c_proj.bias"] = (
            f"{i + 1}.attn.c_proj.bias"
        )
        key_mappings[f"transformer.h.{i}.ln_2.weight"] = (
            f"{i + 1}.mlp.ln.weight"
        )
        key_mappings[f"transformer.h.{i}.ln_2.bias"] = f"{i + 1}.mlp.ln.bias"
        key_mappings[f"transformer.h.{i}.mlp.c_fc.weight"] = (
            f"{i + 1}.mlp.c_fc.weight"
        )
        key_mappings[f"transformer.h.{i}.mlp.c_fc.bias"] = (
            f"{i + 1}.mlp.c_fc.bias"
        )
        key_mappings[f"transformer.h.{i}.mlp.c_proj.weight"] = (
            f"{i + 1}.mlp.c_proj.weight"
        )
        key_mappings[f"transformer.h.{i}.mlp.c_proj.bias"] = (
            f"{i + 1}.mlp.c_proj.bias"
        )
    return key_mappings


def state_dict_from_starcoder(state_dict, config):
    key_mappings = _starcoder_key_mapping_fn(config)
    new_state_dict = {}
    for k, v in state_dict.items():
        new_state_dict[key_mappings[k]] = v

    # split qkv weight and bias
    qdim = config.hidden_dim
    kvdim = config.head_dim * config.n_kv_heads
    for k in ["weight", "bias"]:
        for i in range(config.n_layers):
            if f"{i + 1}.attn.c_attn.linear.{k}" not in new_state_dict:
                continue
            w = new_state_dict[f"{i + 1}.attn.c_attn.linear.{k}"]
            assert w.shape[0] == qdim + kvdim * 2
            new_state_dict[f"{i + 1}.attn.c_attn.q_attn.{k}"] = w[:qdim]
            new_state_dict[f"{i + 1}.attn.c_attn.k_attn.{k}"] = w[
                qdim : qdim + kvdim
            ]
            new_state_dict[f"{i + 1}.attn.c_attn.v_attn.{k}"] = w[
                qdim + kvdim :
            ]
            new_state_dict.pop(f"{i + 1}.attn.c_attn.linear.{k}")
    return new_state_dict


def state_dict_to_starcoder(state_dict, config):
    # merge qkv weight and bias
    qdim = config.hidden_dim
    kvdim = config.head_dim * config.n_kv_heads
    layer_indices = list(set(int(k.split(".")[0]) for k in state_dict.keys()))
    for k in ["weight", "bias"]:
        for i in layer_indices:
            if i == 0 or i == config.n_layers + 1:
                continue
            qw = state_dict[f"{i}.attn.c_attn.q_attn.{k}"]
            kw = state_dict[f"{i}.attn.c_attn.k_attn.{k}"]
            vw = state_dict[f"{i}.attn.c_attn.v_attn.{k}"]
            w = torch.cat([qw, kw, vw], dim=0)
            assert w.shape[0] == qdim + kvdim * 2
            state_dict[f"{i}.attn.c_attn.linear.{k}"] = w
            state_dict.pop(f"{i}.attn.c_attn.q_attn.{k}")
            state_dict.pop(f"{i}.attn.c_attn.k_attn.{k}")
            state_dict.pop(f"{i}.attn.c_attn.v_attn.{k}")

    key_mappings = {v: k for k, v in _starcoder_key_mapping_fn(config).items()}
    new_state_dict = {}
    for k, v in state_dict.items():
        new_state_dict[key_mappings[k]] = v
    return new_state_dict


def starcoder_embedding_layer_names(config: ReaLModelConfig) -> List[str]:
    return ["transformer.wte.weight", "transformer.wpe.weight"]


def starcoder_tblock_param_name(
    config: ReaLModelConfig, layer_idx: int
) -> List[str]:
    names = [
        f"transformer.h.{layer_idx}.ln_1.weight",
        f"transformer.h.{layer_idx}.ln_1.bias",
        f"transformer.h.{layer_idx}.attn.c_attn.weight",
        f"transformer.h.{layer_idx}.attn.c_attn.bias",
        f"transformer.h.{layer_idx}.attn.c_proj.weight",
        f"transformer.h.{layer_idx}.attn.c_proj.bias",
        f"transformer.h.{layer_idx}.ln_2.weight",
        f"transformer.h.{layer_idx}.ln_2.bias",
        f"transformer.h.{layer_idx}.mlp.c_fc.weight",
        f"transformer.h.{layer_idx}.mlp.c_fc.bias",
        f"transformer.h.{layer_idx}.mlp.c_proj.weight",
        f"transformer.h.{layer_idx}.mlp.c_proj.bias",
    ]
    if layer_idx == config.n_layers - 1:
        names += ["transformer.ln_f.weight", "transformer.ln_f.bias"]
    return names


def starcoder_output_head_names(config: ReaLModelConfig) -> List[str]:
    return ["lm_head.weight"]


register_hf_family(
    name="starcoder",
    hf_cls_name="GPTBigCodeForCausalLM",
    config_from_hf_converter=convert_config_starcoder,
    config_to_hf_converter=to_starcoder_config,
    sd_from_hf_converter=state_dict_from_starcoder,
    sd_to_hf_converter=state_dict_to_starcoder,
    embedding_param_names=starcoder_embedding_layer_names,
    tblock_param_names=starcoder_tblock_param_name,
    head_param_names=starcoder_output_head_names,
)



from typing import Callable, Dict
import dataclasses
import gc
import multiprocessing as mp
import os
import queue
import random
import time
import traceback

import pynvml
import torch
import torch.distributed as dist

from realhf.api.core import config
from realhf.api.core.config import ModelFamily
from realhf.base import (
    constants,
    gpu_utils,
    name_resolve,
    namedarray,
    names,
    topology,
)
from realhf.base.topology import ParallelGrid, PipeModelDataParallelTopology

# mp.set_start_method("spawn", force=True)  # Otherwise a CUDA reinitialization error will be thrown

MODEL_NAME = "default"
_DEFAULT_EXPR_NAME = "test"
_DEFAULT_TRIAL_NAME = "test"


class StandaloneTestingProcess(mp.Process):
    """Aims for defining this class:
    + Removing duplicate setup GPU codes in each test.

    Note that `init_global_constants` should be called in `func`.
    """

    def __init__(
        self,
        rank: int,
        world_size: int,
        barrier: mp.Barrier,
        err_queue: mp.Queue,
        func: Callable,
        *args,
        expr_name: str = None,
        trial_name: str = None,
        **kwargs,
    ):
        super().__init__()
        self.rank = rank
        self.world_size = world_size
        self.err_queue = err_queue
        self.barrier = barrier

        self.expr_name = (
            expr_name if expr_name is not None else _DEFAULT_EXPR_NAME
        )
        self.trial_name = (
            trial_name if trial_name is not None else _DEFAULT_TRIAL_NAME
        )

        self.func = func
        self.args = args
        self.kwargs = kwargs

    def _run(self):
        return self.func(*self.args, **self.kwargs)

    def run(self) -> None:
        assert not torch.cuda.is_initialized()
        os.environ["REAL_MODE"] = "LOCAL"
        os.environ["CUDA_DEVICE_MAX_CONNECTIONS"] = "1"

        # isolate cuda devices
        os.environ["CUDA_VISIBLE_DEVICES"] = str(self.rank)
        os.environ["GPU_DEVICES_ISOLATED"] = str(1)
        torch.cuda.set_device(0)

        self.barrier.wait()

        # init process group
        gpu_utils.reveal_ddp_identity(
            self.expr_name, self.trial_name, self.rank
        )
        self.barrier.wait()
        from realhf.impl.model.comm.global_comm import setup_global_comm

        setup_global_comm(self.expr_name, self.trial_name, self.rank)
        # NOTE: The import must be here.
        import deepspeed

        deepspeed.init_distributed()

        # setup some useful constants
        constants.set_experiment_trial_names(self.expr_name, self.trial_name)

        # misc setup
        pynvml.nvmlInit()
        pytorch_memory_burnin(self.rank)

        try:
            self._run()
        except Exception as e:
            print(traceback.format_exc(), flush=True)
            self.err_queue.put(e)
            raise e


def init_global_constants(
    num_dp,
    num_mp,
    num_pp,
    topo=None,
    model_name=None,
    msid2mwid=None,
    sequence_parallel=False,
    gradient_checkpointing=True,
):
    model_name = model_name if model_name is not None else MODEL_NAME

    if topo is None:
        topo = PipeModelDataParallelTopology(
            num_dp=num_dp,
            num_mp=num_mp,
            num_pp=num_pp,
            sequence_parallel=sequence_parallel,
            gradient_checkpointing=gradient_checkpointing,
        )
        ws = num_dp * num_mp * num_pp
    else:
        ws = topo.world_size()

    with constants.model_scope(model_name):
        constants.set_rank_mapping(model_name, topo, msid2mwid=msid2mwid)
        wg_ranks = [constants.to_global_pg_rank(i) for i in range(ws)]
        wg = topology.new_or_get_group(ranks=wg_ranks)

        constants.set_parallelism_group(
            model_name=model_name, pgroup=wg, ranks=wg_ranks
        )
        grid = ParallelGrid(process_group=wg, topology=topo)
        constants.set_grid(model_name=model_name, grid=grid)


class LocalMultiProcessTest:
    """Aims for defining this class:
    1. Defining a barrier and a queue for all sub-processes.
    2. Error handling after launch.
    """

    def __init__(
        self,
        world_size: int,
        func: Callable,
        *args,
        expr_name: str = None,
        trial_name: str = None,
        **kwargs,
    ):
        self.barrier = mp.Barrier(world_size)
        self.err_queue = mp.Queue(world_size)
        self.processes = [
            StandaloneTestingProcess(
                rank,
                world_size,
                self.barrier,
                self.err_queue,
                func,
                *args,
                expr_name=expr_name,
                trial_name=trial_name,
                **kwargs,
            )
            for rank in range(world_size)
        ]

    def launch(self):
        assert not torch.cuda.is_initialized()
        [p.start() for p in self.processes]
        assert not torch.cuda.is_initialized()
        while any([p.is_alive() for p in self.processes]):
            try:
                err = self.err_queue.get_nowait()
                [p.terminate() for p in self.processes]
                raise err
            except queue.Empty:
                time.sleep(0.1)
        [p.join() for p in self.processes]


def clear_name_resolve(expr_name=None, trial_name=None):
    expr_name = expr_name if expr_name is not None else _DEFAULT_EXPR_NAME
    trial_name = trial_name if trial_name is not None else _DEFAULT_TRIAL_NAME
    name_resolve.clear_subtree(
        names.trial_root(experiment_name=expr_name, trial_name=trial_name)
    )


def make_finetune_spec(
    bs_per_device,
    total_train_epochs=1,
    total_train_steps=10,
    steps_per_epoch=10,
    max_seq_len=1024,
):
    import realhf.api.core.model_api as model_api

    finetune_spec = model_api.FinetuneSpec(
        total_train_epochs=total_train_epochs,
        total_train_steps=total_train_steps,
        steps_per_epoch=steps_per_epoch,
    )
    return finetune_spec


def random_sentence(min_len=100, max_len=128):
    words = [
        "the",
        "quick",
        "brown",
        "fox",
        "jumped",
        "over",
        "the",
        "lazy",
        "dog",
    ]
    sentence_length = random.randint(min_len, max_len)
    return " ".join(random.choices(words, k=sentence_length))
    # return "Output less than 50 words:"


def make_input(tokenizer, device, s):
    tokenizer.padding_side = "left"
    prompts = tokenizer(s, return_tensors="pt", padding=True)

    input_ids, attention_mask = prompts["input_ids"], prompts["attention_mask"]
    input_ids = input_ids.to(device)
    attention_mask = attention_mask.to(device)

    # print(f"make input input_ids.shape {input_ids.shape}")

    return input_ids, attention_mask


def make_batch(tokenizer, device, batch_size, dp_rank, dp_worldsize, seed=373):
    random.seed(seed)
    whole_batch = [random_sentence() for _ in range(batch_size)]
    dp_batch = whole_batch[
        batch_size
        // dp_worldsize
        * dp_rank : batch_size
        // dp_worldsize
        * (dp_rank + 1)
    ]
    return make_input(tokenizer, device, dp_batch)


def init_data(tokenizer, device, batch_size, seed, dp_rank=None, num_dp=None):
    from realhf.impl.model.utils.padding import unpad_input

    if dp_rank == None:
        assert num_dp == None
        dp_rank = constants.data_parallel_rank()
        num_dp = constants.data_parallel_world_size()
    input_ids, attention_mask = make_batch(
        tokenizer, device, batch_size, dp_rank % num_dp, num_dp, seed=seed
    )
    packed_input_ids, _, cu_seqlens, max_seqlen = unpad_input(
        input_ids, attention_mask
    )
    prompt_mask = torch.zeros_like(packed_input_ids)
    data = namedarray.NamedArray(
        packed_input_ids=packed_input_ids,
        cu_seqlens=cu_seqlens,
        prompts=input_ids,
        prompt_mask=prompt_mask.bool(),
        prompt_att_mask=attention_mask,
    )
    return data


def random_sample(bs, seq_len, vocab_size):
    import torch

    from realhf.impl.model.utils.padding import unpad_input

    input_ids = torch.randint(0, vocab_size, (bs, seq_len), dtype=torch.long)
    attention_mask = torch.ones_like(input_ids)
    packed_input_ids, _, cu_seqlens, max_seqlen = unpad_input(
        input_ids, attention_mask
    )
    prompt_mask = torch.zeros_like(packed_input_ids)
    data = namedarray.NamedArray(
        packed_input_ids=packed_input_ids,
        cu_seqlens=cu_seqlens,
        prompts=input_ids,
        prompt_mask=prompt_mask.bool(),
        prompt_att_mask=attention_mask,
    )
    return data


def pytorch_memory_burnin(rank):
    torch.cuda.set_device(0)
    torch.cuda.init()
    x = torch.randn(1, device="cuda", dtype=torch.float64, requires_grad=True)
    y = x * torch.randn(1000, device="cuda", dtype=torch.float64)
    y.mean().backward()
    del x, y
    gc.collect()
    torch.cuda.empty_cache()
    gc.collect()


def clear_gpu_cache():
    gc.collect()
    torch.cuda.empty_cache()
    gc.collect()


def get_memory(rank):
    handle = pynvml.nvmlDeviceGetHandleByIndex(rank)
    memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    # total_memory = memory_info.total / (1024**2)  # Convert bytes to megabytes
    used_memory = memory_info.used / (1024**2)
    return used_memory


def get_pytorch_profiler(save_fn: str):

    def trace_handler(p: torch.profiler._KinetoProfile):
        # print(
        #     p.key_averages().table(
        #         sort_by="cuda_memory_usage", row_limit=20, max_name_column_width=30, max_src_column_width=30
        #     )
        # )
        p.export_chrome_trace(save_fn)

    return torch.profiler.profile(
        activities=[
            torch.profiler.ProfilerActivity.CPU,
            torch.profiler.ProfilerActivity.CUDA,
        ],
        record_shapes=True,
        profile_memory=True,
        with_stack=True,
        on_trace_ready=trace_handler,
        with_flops=True,
    )


def get_llama_config(size):
    from realhf.api.core.model_api import ReaLModelConfig

    if size == 7:
        size_args = dict(
            n_layers=40,
            n_kv_heads=32,
            head_dim=128,
            hidden_dim=4096,
            intermediate_dim=11008,
            n_positions=4096,
        )
    elif size == 13:
        size_args = dict(
            n_layers=40,
            n_kv_heads=40,
            head_dim=128,
            hidden_dim=5120,
            intermediate_dim=13824,
            n_positions=4096,
        )
    elif size == 34:
        size_args = dict(
            n_layers=48,
            n_kv_heads=8,
            head_dim=128,
            hidden_dim=8192,
            intermediate_dim=22016,
            n_positions=16384,
        )
    elif size == 70:
        size_args = dict(
            n_layers=80,
            n_kv_heads=8,
            head_dim=128,
            hidden_dim=8192,
            intermediate_dim=28672,
            n_positions=32768,
        )
    else:
        raise ValueError(f"size {size} not supported")

    return ReaLModelConfig(
        vocab_size=32000,
        activation_function="silu",
        use_attention_bias=False,
        layer_norm_type="rms",
        mlp_type="llama",
        apply_rotary=True,
        **size_args,
    )
